# APIs for Windows 32-bit MSVC 2008 runtime library (msvcr90).
# Format:  rettype, retname, callconv, exactname, arglist(type, name)
#          arglist type is one of ['int', 'void *']
#          arglist name is one of [None, 'funcptr', 'obj', 'ptr']
#
# The subset of functions defined below accounts for 80% of observed calls.

api_defs = {
    'msvcr90.??2@yapaxi@z':( 'int', None, 'cdecl', 'msvcr90.??2@YAPAXI@Z', (('int', None),) ),
    'msvcr90.??_u@yapaxi@z':( 'int', None, 'cdecl', 'msvcr90.??_U@YAPAXI@Z', (('int', None),) ),
    'msvcr90.??3@yaxpax@z':( 'void', None, 'cdecl', 'msvcr90.??3@YAXPAX@Z', (('void *', 'ptr'),) ),
    'msvcr90.??_v@yaxpax@z':( 'void', None, 'cdecl', 'msvcr90.??_V@YAXPAX@Z', (('void *', 'ptr'),) ),

    'msvcr90._cxxthrowexception':( 'void', None, 'stdcall', 'msvcr90._CxxThrowException', (('void *', 'ptr'), ('int', None)) ),
    'msvcr90.__iob_func':( 'int', None, 'cdecl', 'msvcr90.__iob_func', () ),
    'msvcr90._invalid_parameter_noinfo':( 'void', None, 'cdecl', 'msvcr90._invalid_parameter_noinfo', () ),
    'msvcr90._encode_pointer':( 'int', None, 'cdecl', 'msvcr90._encode_pointer', (('int', None),) ),
    'msvcr90._errno':( 'int', None, 'cdecl', 'msvcr90._errno', () ),

    'msvcr90._exit':( 'int', None, 'cdecl', 'msvcr90._exit', (('int', None),) ),
    'msvcr90.abort':( 'int', None, 'cdecl', 'msvcr90.abort', () ),
    'msvcr90.exit':( 'int', None, 'cdecl', 'msvcr90.exit', (('int', None),) ),
    'msvcr90.quick_exit':( 'int', None, 'cdecl', 'msvcr90.quick_exit', (('int', None),) ),

    'msvcr90.calloc':( 'int', None, 'cdecl', 'msvcr90.calloc', (('int', None), ('int', None)) ),
    'msvcr90.malloc':( 'int', None, 'cdecl', 'msvcr90.malloc', (('int', None),) ),
    'msvcr90.free':( 'void', None, 'cdecl', 'msvcr90.free', (('void *', 'ptr'),) ),

    'msvcr90.memcmp':( 'int', None, 'cdecl', 'msvcr90.memcmp', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'msvcr90.memcpy':( 'int', None, 'cdecl', 'msvcr90.memcpy', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'msvcr90.memmove':( 'int', None, 'cdecl', 'msvcr90.memmove', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'msvcr90.memmove_s':( 'int', None, 'cdecl', 'msvcr90.memmove_s', (('void *', 'ptr'), ('int', None), ('void *', 'ptr'), ('int', None)) ),
    'msvcr90.memset':( 'int', None, 'cdecl', 'msvcr90.memset', (('void *', 'ptr'), ('int', None), ('int', None)) ),

    'msvcr90._snprintf':( 'int', None, 'cdecl', 'msvcr90._snprintf', (('int', None), ('int', None), ('int', None)) ),
    'msvcr90._swprintf':( 'int', None, 'cdecl', 'msvcr90._snprintf', (('int', None), ('int', None), ('int', None), ('int', None)) ),
    'msvcr90.fprintf':( 'int', None, 'cdecl', 'msvcr90.fprintf', (('void *', 'ptr'), ('int', None)) ),
    'msvcr90.sprintf':( 'int', None, 'cdecl', 'msvcr90.sprintf', (('int', None), ('int', None)) ),
    'msvcr90.printf':( 'int', None, 'cdecl', 'msvcr90.printf', (('int', None),) ),

    'msvcr90._stricmp':( 'int', None, 'cdecl', 'msvcr90._stricmp', (('void *', 'ptr'), ('void *', 'ptr')) ),
    'msvcr90.strchr':( 'int', None, 'cdecl', 'msvcr90.strchr', (('void *', 'ptr'), ('int', None)) ),
    'msvcr90.strcmp':( 'int', None, 'cdecl', 'msvcr90.strcmp', (('void *', 'ptr'), ('void *', 'ptr')) ),
    'msvcr90.strcpy':( 'int', None, 'cdecl', 'msvcr90.strcpy', (('void *', 'ptr'), ('void *', 'ptr')) ),
    'msvcr90.strlen':( 'int', None, 'cdecl', 'msvcr90.strlen', (('void *', 'ptr'),) ),
    'msvcr90.strncmp':( 'int', None, 'cdecl', 'msvcr90.strncmp', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'msvcr90.strncpy':( 'int', None, 'cdecl', 'msvcr90.strncpy', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None)) ),
    'msvcr90.strstr':( 'int', None, 'cdecl', 'msvcr90.strstr', (('void *', 'ptr'), ('void *', 'ptr')) ),
    'msvcr90._wcsicmp':( 'int', None, 'cdecl', 'msvcr90._wcsicmp', (('void *', 'ptr'), ('void *', 'ptr')) ),
    'msvcr90.wcslen':( 'int', None, 'cdecl', 'msvcr90.wcslen', (('void *', 'ptr'),) ),
    'msvcr90.wcsrchr':( 'int', None, 'cdecl', 'msvcr90.wcsrchr', (('void *', 'ptr'), ('int', None)) ),

    'msvcr90._time64':( 'int', None, 'cdecl', 'msvcr90._time64', (('void *', 'ptr'),) ),
    'msvcr90.atoi':( 'int', None, 'cdecl', 'msvcr90.atoi', (('void *', 'ptr'),) ),
    'msvcr90.getenv':( 'int', None, 'cdecl', 'msvcr90.getenv', (('void *', 'ptr'),) ),
    'msvcr90.fclose':( 'int', None, 'cdecl', 'msvcr90.fclose', (('void *', 'ptr'),) ),
    'msvcr90.bsearch':( 'int', None, 'cdecl', 'msvcr90.bsearch', (('void *', 'ptr'), ('void *', 'ptr'), ('int', None), ('int', None), ('void *', 'funcptr')) )
}



import sys
import logging

import envi
import vivisect.const as v_const
import vivisect.impemu.monitor as viv_imp_monitor

logger = logging.getLogger(__name__)


def getMainVas(vw):
    ret = []
    for va, name in vw.getNames():
        lcsm = '__libc_start_main_%.8x' % va
        if name in (lcsm, "*."+lcsm):
            ret.append(va)

            # also grab any function wrappers
            for xfr, xto, xtype, xinfo in vw.getXrefsTo(va, rtype=v_const.REF_CODE):
                funcva = vw.getFunction(xfr)
                if not vw.isFunctionThunk(funcva):
                    continue
                ret.append(xfr)
    return ret


def analyzeFunction(vw, funcva, lcsm=None):
    '''
    search through all calls, looking for a call to an import named __libc_start_main
    then check for arg0
    '''
    try:
        emu = vw.getEmulator(va=funcva)
        if lcsm is None:
            lcsm = getMainVas(vw)
        emumon = AnalysisMonitor(vw, funcva, lcsm)

        emu.setEmulationMonitor(emumon)
        emu.runFunction(funcva, maxhit=1)

        if not emumon.success:
            logger.info("  emumon failure: %r", vars(emumon))
            return
        try:
            api = vw.getFunctionApi(emumon.startmain)
            cconv = emu.getCallingConvention(api[vivisect.API_CCONV])
        except:
            ccname = vw.getMeta('DefaultCall')
            cconv = emu.getCallingConvention(ccname)

        args = cconv.getPreCallArgs(emu, 1)
        mainva = args[0]

        vw.addEntryPoint(mainva)
        logger.debug('discovered new function: 0x%x', mainva)
        vw.makeFunction(mainva)

        curname = vw.getName(mainva)
        if curname in (None, "sub_%.8x" % mainva):
            vw.makeName(mainva, 'main', True)

    except Exception as e:
        sys.excepthook(*sys.exc_info())


def analyze(vw):
    mainVas = getMainVas(vw)
    for va in mainVas:
        for xfr, xto, xtype, xinfo in vw.getXrefsTo(va, rtype=v_const.REF_CODE):
            logger.info("0x%x -> 0x%x (%r)", xfr, xto, xtype)
            funcva = vw.getFunction(xfr)
            if funcva is None:
                continue

            analyzeFunction(vw, funcva, lcsm=mainVas)


class AnalysisMonitor(viv_imp_monitor.AnalysisMonitor):

    def __init__(self, vw, fva, lcsm):
        viv_imp_monitor.AnalysisMonitor.__init__(self, vw, fva)
        self.success = False
        self.emu = None
        self.startmain = lcsm

    def prehook(self, emu, op, starteip):

        if op.iflags & envi.IF_CALL:
            # it's a call, get the target
            branches = [br for br in op.getBranches(emu) if not (br[1] & envi.BR_FALL)]
            logger.debug('libc_start_main: %r\t\ttgts: %r', [hex(x) for x in self.startmain], branches)

            # check if it matches what we believe to be __libc_start_main
            for branch in branches:
                tgt, flags = branch
                if tgt in self.startmain:
                    self.success = True
                    self.emu = emu
                    emu.stopEmu()




import os
import re
import time
import socket
import struct
import logging
import binascii
import tempfile
import threading

import PE
import envi
import vtrace

import envi.bits as e_bits
import envi.common as e_common
import envi.registers as e_registers
import vtrace.platforms.base as v_base
import envi.symstore.resolver as e_resolv

'''
VMWare config options...
debugStub.listen.guest64 = "TRUE" # ends up on port 8864 (or next avail)
debugStub.listen.guest32 = "TRUE" # ....            8832

debugStub.listen.guest32.remote = "TRUE"

debugStub.hideBreakpoints = "TRUE" # Enable breakpoints

From GDB stuff...
===== i386
src/gdb/i386-tdep.c
src/gdb/regformats/reg-i386.dat
===== amd64
src/gdb/amd64-tdep.c
src/gdb/regformats/reg-x86-64.dat
===== arm
name:arm
expedite:r11,sp,pc
32:r0
32:r1
32:r2
32:r3
32:r4
32:r5
32:r6
32:r7
32:r8
32:r9
32:r10
32:r11
32:r12
32:sp
32:lr
32:pc
96:f0
96:f1
96:f2
96:f3
96:f4
96:f5
96:f6
96:f7
32:fps
32:cpsr
'''

logger = logging.getLogger(__name__)

gdb_reg_defs = {
    'i386': (
        ['eax','ecx','edx','ebx','esp','ebp','esi','edi','eip','eflags','cs','ss','ds','es','fs','gs'
        ],
        '<16I'
    ),

    'amd64': (
        ['rax','rbx','rcx','rdx','rsi','rdi','rbp','rsp',
         'r8','r9','r10','r11','r12','r13','r14','r15','rip',
         'eflags','cs','ss','ds','es','fs','gs',
         #'st0','st1','st2','st3','st4','st5','st6','st7',
         #'fctrl','fstat','ftag','fiseg','fioff','foseg','fooff','fop'
        ],

        #'<17Q7L' + ('10s' * 8) + '8L'
        '<17Q7L'
    ),

    # FIXME we will need arm flavors...
    'arm': (
        ["r0","r1","r2","r3","r4","r5","r6","r7","r8","r9","sl","fp","ip","sp",
         "lr","pc", None, "cpsr"],
        '<16I96sI'
    ),
}

exit_types = ('X', 'W')

def pkt(cmd):
    return '$%s#%.2x' % (cmd, csum(cmd))

def csum(bytes):
    sum = 0
    for b in bytes:
        sum += ord(b)
    return sum & 0xff

SIGINT  = 2
SIGTRAP = 5

trap_sigs = (SIGINT, SIGTRAP)

class GdbServerDisconnected(Exception):
    pass

class GdbStubMixin:

    def __init__(self, host=None, port=None):
        self._gdb_host = host
        self._gdb_port = port
        self._gdb_sock = None

        self._gdb_tx_lock = threading.Lock()  # socket tx lock
        self._gdb_rx_lock = threading.Lock()  # socket rx lock
        self._gdb_tns_lock = threading.Lock() # for "transact" API

        self._gdb_filemagic = None # Tracers may use this to trigger _findLibraryMaps

        # These get set by _gdbSetRegisterInfo
        self._gdb_regfmt = ''
        self._gdb_regsize = 0
        self._gdb_reg_xlat = []
        #self._gdb_regnames = []

        self.stepping = False
        self.breaking = False

        self._gdbSetArch( self.getMeta('Architecture') )

    def _recvUntil(self, c):
        ret = ''
        while not ret.endswith(c):
            x = self._gdb_sock.recv(1)
            if len(x) == 0:
                raise Exception('socket closed prematurely!')
            ret += x
        return ret

    def _recvPkt(self):

        with self._gdb_rx_lock:

            b = self._gdb_sock.recv(1)
            if len(b) == 0:
                raise GdbServerDisconnected()

            if b != '$':
                raise Exception('Invalid Pkt Beginning! ->%s<-' % b)

            bytes = self._recvUntil('#')
            bytes = bytes[:-1]

            isum = int(self._gdb_sock.recv(2), 16)
            ssum = csum(bytes)
            if isum != ssum:
                raise Exception('Invalid Checksum! his: 0x%.2x ours: 0x%.2x' % (isum, ssum))

            self._gdb_sock.sendall('+')

            logger.debug('RECV: ->%s<-', bytes)
            return bytes

    def _gdbAddMemBreak(self, addr, size):
        resp = self._cmdTransact('Z0,%x,%x' % (addr, size))
        self._raiseIfError( resp )

    def _gdbDelMemBreak(self, addr, size):
        resp = self._cmdTransact('z0,%x,%x' % (addr, size))
        self._raiseIfError( resp )

    def _cmdTransact(self, cmd):
        with self._gdb_tns_lock:
            self._sendPkt(cmd)
            return self._recvPkt()

    def _sendPkt(self, cmd):
        logger.debug('SEND: ->%s<-', cmd)
        with self._gdb_tx_lock:

            self._gdb_sock.sendall(pkt(cmd))
            b = self._gdb_sock.recv(1)
            if b != '+':
                raise Exception('Retrans! ->%s<-' % b)

    def _connectSocket(self):
        if self._gdb_sock is not None:
            self._gdb_sock.shutdown(2)

        tries = 0
        while tries < 10:
            self._gdb_sock = socket.socket()
            try:
                self._gdb_sock.connect( (self._gdb_host, self._gdb_port) )

                # Some gdb stubs seem to send/expect an initial '+'
                try:
                    self._gdb_sock.settimeout(1)
                    self._gdb_sock.recv(1)
                    self._gdb_sock.sendall('+')

                except socket.timeout:
                    pass

                self._gdb_sock.settimeout(None)
                break

            except Exception:
                time.sleep(0.2)
                tries += 1

    def _monitorCommand(self, cmd):
        resp = ''
        cmd = 'qRcmd,%s' % e_common.hexify(cmd)
        pkt = self._cmdTransact(cmd)
        while not pkt.startswith('OK'):
            self._raiseIfError(pkt)
            if not pkt.startswith('O'):
                return binascii.unhexlify(pkt)
            resp += binascii.unhexlify(pkt[1:])
            pkt = self._recvPkt()
        return resp

    def platformAttach(self, pid):
        self._connectSocket()
        self.attaching = True
        # Wait for the debug stub to stop the target
        while True:
            pkt = self._cmdTransact('?')
            if len(pkt) == 0:
                raise Exception('Attach Response Error!')

            if int(pkt[1:3], 16) == 0:
                import time
                time.sleep(0.1)
                self.platformSendBreak()
                pkt = self._cmdTransact('?')
            break
        self._sendPkt('?')

    def platformContinue(self):
        sig = self.getCurrentSignal()
        cmd = 'c'
        if sig is not None:
            cmd = 'C%.2x' % sig
        self._sendPkt(cmd)

    def platformStepi(self):
        # FIXME by selected thread? and address?
        #self._cmdTransact('s')
        self._sendPkt('s')
        self.stepping = True

    def platformDetach(self):
        if not self.running:
            self.platformContinue()
        self._gdb_sock.shutdown(2)
        self._gdb_sock = None

    def platformSendBreak(self):
        '''
        For now, the only way I know how to re-break the target
        is to disconnect and re-connect...  TOTALLY GHETTO HACK!
        '''
        # If this isn't a break during attach, tell everybody we are
        # breaking...
        if not self.attaching:
            self.breaking = True
        self._gdb_sock.sendall('\x03')

    def platformWait(self):
        while True:
            pkt = self._recvPkt()
            if pkt.startswith('O'):
                continue
            break
        return pkt

    def _gdbJustAttached(self):
        pass

    def _gdbCreateThreads(self):
        self._simpleCreateThreads()

    def _gdbLoadLibraries(self):
        if self._gdb_filemagic:
            self._findLibraryMaps(self._gdb_filemagic, always=True)

    def platformProcessEvent(self, event):
        logger.debug('EVENT ->%s<-', str(event))

        if len(event) == 0:
            self.setMeta('ExitCode', 0xffffffff)
            self.fireNotifiers(vtrace.NOTIFY_EXIT)
            self._gdb_sock.shutdown(2)
            self._gdb_sock = None
            return

        atype = event[0]
        signo = int(event[1:3], 16)

        # Is this a thread specific signal?
        if atype == 'T':

            logger.debug('Signal: %s', str(sig))

            dictbytes = event[3:]

            evdict = {}
            for kvstr in dictbytes.split(';'):
                if not kvstr:
                    break
                key, value = kvstr.split(':', 1)
                evdict[key.lower()] = value

            # Did we get a specific thread?
            tidstr = evdict.get('thread')
            if tidstr is not None:
                tid = int(tidstr, 16)
                self.setMeta('ThreadId', tid)
            else:
                logger.warning("We should ask for the current thread here!")

        elif atype == 'S':
            pass

        elif atype in exit_types:

            # Fire an exit event and GTFO!
            self._fireExit(signo)
            return

        else:
            logger.warning('Unhandled Gdb Server Event: %s', str(event))

        # if self.attaching and signo in trap_sigs:
        if self.attaching:
            self.attaching = False

            self._gdbJustAttached()
            self._gdbLoadLibraries()
            self._gdbCreateThreads()

            self.runAgain(False)  # Clear this, if they want BREAK to run, it will
            self.fireNotifiers(vtrace.NOTIFY_BREAK)

        elif self.breaking and signo in trap_sigs:
            self.breaking = False
            self.fireNotifiers(vtrace.NOTIFY_BREAK)

        # Process the signal and decide what to do...
        elif signo == SIGTRAP:

            # Traps on posix systems are a little complicated
            if self.stepping:
                #FIXME try out was single step thing for intel
                self.stepping = False
                self.fireNotifiers(vtrace.NOTIFY_STEP)

            elif self.checkBreakpoints():
                return

            #elif self.checkWatchpoints():
                #return

            #elif self.checkBreakpoints():
                # It was either a known BP or a sendBreak()
                #return

            #elif self.execing:
                ##self.execing = False
                #self.handleAttach()

            else:
                self._fireSignal(signo)

        #elif signo == signal.SIGSTOP:
            #self.handleAttach()

        else:
            self._fireSignal(signo)

    #def _gdbCreateThreads(self):
        #initid = self.getMeta('ThreadId')
        #for tid in self.platformGetThreads().keys():
            #self.setMeta('ThreadId', tid)
            #self.fireNotifiers(vtrace.NOTIFY_CREATE_THREAD)
        #self.setMeta('ThreadId', initid)

    #def _gdbSetRegisterInfo(self, fmt, names):
        # Used by the Trace implementations to tell the gdb
        # stub code how to unpack the register buf

        #self._gdb_regfmt = fmt
        #self._gdb_regnames = names

        #self._gdb_reg_xlat = []
        #self._gdb_regsize = struct.calcsize(fmt)


        #for i,name in enumerate(names):
            #if name is None: # So we can skip parts of the gdb definition...
                #continue
            #j = self.getRegisterIndex(name)
            #if j is not None:
                #self._gdb_reg_xlat.append( (i, j) )

    def _gdbSetArch(self, arch):

        reginfo = gdb_reg_defs.get(arch)
        if reginfo is None:
            raise Exception('No known register mappings for gdbstub on %s!' % arch)

        regnames, regfmt = reginfo

        self._gdb_regfmt = regfmt
        self._gdb_regsize = struct.calcsize( regfmt )
        self._gdb_regnames = regnames

    def platformGetRegCtx(self, tid):
        '''
        Get an envi register context from the target stub.
        '''

        # FIXME tid!
        regbuf = self._cmdTransact('g')
        regbytes = self._runLengthDecode(regbuf)
        rvals = struct.unpack(self._gdb_regfmt, regbytes[:self._gdb_regsize])
        ctx = self.arch.archGetRegCtx()

        for i,regval in enumerate(rvals):
            regname = self._gdb_regnames[i]
            if regname is not None:
                ctx.setRegisterByName( regname, regval )

        return ctx

    def platformSetRegCtx(self, tid, ctx):
        '''
        Set the target stub's register context from the envi register context
        '''
        # FIXME tid!
        regbytes = binascii.unhexlify(self._cmdTransact('g'))
        regremain = regbytes[self._gdb_regsize:]
        rvals = struct.unpack(self._gdb_regfmt, regbytes[:self._gdb_regsize])
        rvals = list(rvals) # So we can assign to them...
        for myidx, enviidx in self._gdb_reg_xlat:
            rvals[myidx] = ctx.getRegister(enviidx)
        newbytes = struct.pack(self._gdb_regfmt, rvals) + regremain
        return self._cmdTransact('G' + e_common.hexify(newbytes))

    def platformGetThreads(self):

        ret = {}

        self._sendPkt('qfThreadInfo')
        tbytes = self._recvPkt()

        while tbytes.startswith('m'):

            if tbytes.find(','):
                for bval in tbytes[1:].split(','):
                    ret[int(bval, 16)] = 0
            else:
                ret[int(tbytes[1:], 16)] = 0

            self._sendPkt('qsThreadInfo')
            tbytes = self._recvPkt()

        return ret

    def _raiseIfError(self, msg):
        if msg.startswith('E'):
            raise Exception('Error: %s' % msg)

    def _runLengthDecode(self, buf):
        # GDB RSP implements some run-length encoding to save space
        i = buf.find('*')
        while i != -1:
            cnt = ord(buf[i+1]) - 29 # Run-length encoding is minus 29...
            pad = buf[i-1] * cnt
            buf = buf[:i] + pad + buf[i+2:]

            i = buf.find('*')

        return binascii.unhexlify(buf)

    def platformReadMemory(self, addr, size):
        mbytes = ''
        offset = 0
        logger.debug('READ: 0x%.8x (%d)', addr, size)
        while len(mbytes) < size:
            # FIXME is this 256 problem just in the VMWare gdb stub?
            cmd = 'm%x,%x' % (addr + offset, min(256, size-offset))
            pkt = self._cmdTransact(cmd)
            self._raiseIfError(pkt)
            pbytes = self._runLengthDecode(pkt)
            offset += len(pbytes)
            mbytes += pbytes
        return mbytes

    def platformWriteMemory(self, addr, mbytes):
        cmd = 'M%x,%x:%s' % (addr, len(mbytes), e_common.hexify(mbytes))
        pkt = self._cmdTransact(cmd)

    def platformGetMaps(self):
        # No way to enumerate these by default...
        return []

    def platformPs(self):
        return [ (1, 'GdbStubTarget'), ]

    def platformGetFds(self):
        return []

# FROM HERE DOWN IS ALL CRAP THAT IS STILL GETTING SORTED OUT

class GdbStubMixin_old(e_registers.RegisterContext):

    def __init__(self):

        self.stepping = False
        self.attaching = False
        self.breaking = False

        self.bigmask = e_bits.u_maxes[ self.getPointerSize() ]

        aname = self.getMeta('Architecture')
        self._addArchNamespace(aname)

        self.setMeta('Platform', 'gdbstub')

        self.setMeta('GdbServerHost', 'localhost')
        self.setMeta('GdbServerPort', 0)
        self.setMeta('GdbPlatform', 'Unknown')
        self.setMeta('GdbTargetPlatform', 'Unknown')

        self.setMeta('BinaryFormat', None)

        arch_reg_info = gdb_reg_defs.get(aname)
        if arch_reg_info is None:
            raise Exception('We dont know the GDB register definition for arch: %s' % name)

        self._arch_regnames, self._arch_regfmt = arch_reg_info
        self._arch_regsize = struct.calcsize(self._arch_regfmt)

        self._arch_rctx = self.arch.archGetRegCtx()
        self._arch_reg_xlat = []
        for i,name in enumerate(self._arch_regnames):
            if name is None: # So we can skip parts of the gdb definition...
                continue
            j = self._arch_rctx.getRegisterIndex(name)
            if j is not None:
                self._arch_reg_xlat.append((i,j))

        # Load up our register definition!
        e_registers.RegisterContext.__init__(self)
        rinfo = self._arch_rctx.getRegisterInfo(meta=True)
        self.setRegisterInfo(rinfo)

    def _addArchNamespace(self, aname):
        if aname == 'arm':
            import vstruct.defs.arm7 as vs_arm7
            self.vsbuilder.addVStructNamespace('arm7', vs_arm7)

    def normFileName(self, fname):
        # We don't know if it's / or \ ...   do both!
        basename = fname.split('/')[-1].split('\\')[-1]
        return basename.split(".")[0].split("-")[0].lower()

    def platformParseBinary(self, filename, baseaddr, normname):
        logger.warning('Not implemented: platformParseBinary: 0x%.8x %s', baseaddr, normname)

    def platformParseBinaryPe(self, filename, baseaddr, normname):

        # If we're on windows, fake out the PE header and use dbghelp
        if False:
            # FIXME this code is stolen and should be a function!
            import ctypes
            import vtrace.platforms.win32 as vt_win32
            fakepe = self.readMemory(baseaddr, 1024)
            tfile = tempfile.NamedTemporaryFile(delete=False)
            tfilename = tfile.name
            pebuf = ctypes.create_string_buffer(fakepe)
            try:
                try:
                    tfile.write(fakepe)
                    tfile.close()
                    #parser = vt_win32.Win32SymbolParser(-1, tfilename, baseaddr)
                    parser = vt_win32.Win32SymbolParser(-1, None, ctypes.addressof(pebuf))
                    parser.parse()
                    parser.loadSymsIntoTrace(self, normname)
                finally:
                    os.unlink(tfilename)
            except Exception as e:
                logger.warning(str(e))

        else:
            pe = PE.peFromMemoryObject(self, baseaddr)
            for rva, ord, name in pe.getExports():
                self.addSymbol(e_resolv.Symbol(name, baseaddr+rva, 0, normname))

    def platformPs(self):
        return [ (1, 'SystemProcess'), ]

    def _getVmwareReg(self, rname):
        '''
        Use VMWare's monitor extension to get a register we wouldn't
        normally have...
        '''
        #fs 0x30 base 0xffdff000 limit 0x00001fff type 0x3 s 1 dpl 0 p 1 db 1
        fsstr = self._monitorCommand('r %s' % rname)
        fsparts = fsstr.split()
        return int(fsparts[3], 16)

    def _getVmwareIdtr(self):
        istr = self._monitorCommand('r idtr')
        m = re.match('.* base=(0x\w+) .*', istr)
        idtr = int(m.groups()[0], 0)
        return idtr

    def _getNtOsKrnl(self, idtr):
        x1, kptr, x2 = self.readMemoryFormat(idtr, '<IQI')
        try:
            kptr -= kptr & 0xfff
            while not self.readMemory(kptr, 16).startswith('MZ\x90\x00'):
                kptr -= 4096
            return kptr
        except Exception as e:
            return None

    def _enumTargetOs(self, fsbase):

        self.setVariable('fsbase', fsbase)

        fs_fields = self.readMemoryFormat(fsbase, '<8I')

        # Windows has a self reference in the KPCR...
        if fs_fields[7] == fsbase:

            # Use KPCR from XP for now...
            import vstruct.defs.windows.win_5_1_i386.ntoskrnl as vs_w_ntoskrnl
            self.vsbuilder.addVStructNamespace('nt', vs_w_ntoskrnl)

            self.setMeta('GdbTargetPlatform', 'windows')
            self.casesens = False

            kpcr = self.getStruct('nt.KPCR', fsbase)
            kver = self.getStruct('nt.DBGKD_GET_VERSION64', kpcr.KdVersionBlock)

            kernbase = kver.KernBase & self.bigmask
            modlist = kver.PsLoadedModuleList & self.bigmask

            self.setVariable('PsLoadedModuleList', modlist)
            self.setVariable('KernelBase', kernbase)

            self.platformParseBinary = self.platformParseBinaryPe

            self.fireNotifiers(vtrace.NOTIFY_ATTACH)

            self.addLibraryBase('nt', kernbase, always=True)
            ldr_entry = self.readMemoryFormat(modlist, '<I')[0]
            while ldr_entry != modlist:
                ldte = self.getStruct('nt.LDR_DATA_TABLE_ENTRY', ldr_entry)
                dllname = self.readMemory(ldte.FullDllName.Buffer, ldte.FullDllName.Length).decode('utf-16le')
                dllbase = ldte.DllBase & self.bigmask
                self.addLibraryBase(dllname, dllbase, always=True)
                ldr_entry = ldte.InLoadOrderLinks.Flink & self.bigmask


        else:
            # FIXME enumerate non-windows OSs!
            self.fireNotifiers(vtrace.NOTIFY_ATTACH)

    def _enumGdbTarget(self):
        psize = self.getPointerSize()
        vercmd = self._monitorCommand('version')

        monhelp = self._monitorCommand('help')

        if monhelp.find('netdev_add') != -1:

            self.setMeta('GdbPlatform', 'Qemu32')

            fsbase = None
            monreg = self._monitorCommand('info registers')
            for line in monreg.split('\n'):
                if not line.startswith('FS'):
                    continue
                parts = line.split()
                fsbase = int(parts[2], 16)
                break

            self._enumTargetOs(fsbase)
            #logger.debug(monreg)
            #m = re.match('FS =\w+ (\w+)', monreg, re.G)
            #fsbase = long(m.groups()[0], 0)
            #logger.debug('FSBASE',hex(fsbase))

        elif monhelp.find('linuxoffsets') != -1:

            self.setMeta('GdbPlatform', 'VMware%d' % (psize * 8))

            if psize == 4: # Use the fs register to get KPCR
                fsbase = self._getVmwareReg('fs')
                self._enumTargetOs(fsbase)

            else: # FIXME 64bit vmware!

                idtr = self._getVmwareIdtr()
                self.setVariable('idtr', idtr)

                win_kpcr = 0x07fffffde000

                fields = [-1,]
                try:
                    fields = self.readMemoryFormat(win_kpcr, '<7Q')
                except Exception as e:
                    logger.warning(str(e))

                # FIXME other heuristics for linux/bsd/etc...
                if fields[-1] == win_kpcr:
                    self._initWin64(win_kpcr)
                else:
                    self.fireNotifiers(vtrace.NOTIFY_ATTACH)

                #fsbase = self._getVmwareReg('fs')
                #self.setVariable('fsbase', fsbase)

                #fs_fields = self.readMemoryFormat(fsbase, '<8I')

                #nt = self._getNtOsKrnl(idtr)
                #if nt is not None:
                    # We are 64bit windows!
                    #import vstruct.defs.windows.win_6_1_amd64.ntoskrnl as vs_w_ntoskrnl
                    #self.vsbuilder.addVStructNamespace('nt', vs_w_ntoskrnl)
                    #self.setMeta('GdbTargetPlatform', 'windows')
                    #self.setVariable('KernelBase', nt)
                    #self.platformParseBinary = self.platformParseBinaryPe
                    #self.fireNotifiers(vtrace.NOTIFY_ATTACH)
                    #self.addLibraryBase('nt', nt, always=True)

                #else:

        elif vercmd.lower().find('open on-chip debugger') != -1:

            self.setMeta('GdbPlatform', 'OpenOCD')
            self.fireNotifiers(vtrace.NOTIFY_ATTACH)

        else:
            logger.warning('Unidentified gdbstub: %s', vercmd)
            self.fireNotifiers(vtrace.NOTIFY_ATTACH)


    # FIXME implement getRegister(idx) and steal get/set for regs which are not part of the whole...
    def _initWin64(self, kpcr):

        import vstrct.defs.windows.win_6_1_amd64.ntoskrnl as vs_w_ntoskrnl
        self.vsbuilder.addVStructNamespace('nt', vs_w_ntoskrnl)
        self._initWinBase()
                #nt = self._getNtOsKrnl(idtr)
                #if nt is not None:
                    # We are 64bit windows!
                    #import vstruct.defs.windows.win_6_1_amd64.ntoskrnl as vs_w_ntoskrnl
                    #self.vsbuilder.addVStructNamespace('nt', vs_w_ntoskrnl)
                    #self.setMeta('GdbTargetPlatform', 'windows')
                    #self.setVariable('KernelBase', nt)
                    #self.platformParseBinary = self.platformParseBinaryPe
                    #self.fireNotifiers(vtrace.NOTIFY_ATTACH)
                    #self.addLibraryBase('nt', nt, always=True)

    def _initWinBase(self, kpcr):

        self.setMeta('GdbTargetPlatform', 'windows')
        self.casesens = False

        kpcr = self.getStruct('nt.KPCR', kpcr)
        kver = self.getStruct('nt.DBGKD_GET_VERSION64', kpcr.KdVersionBlock)

        kernbase = kver.KernBase & self.bigmask
        modlist = kver.PsLoadedModuleList & self.bigmask

        self.setVariable('PsLoadedModuleList', modlist)
        self.setVariable('KernelBase', kernbase)

        self.platformParseBinary = self.platformParseBinaryPe

        self.fireNotifiers(vtrace.NOTIFY_ATTACH)

        self.addLibraryBase('nt', kernbase, always=True)
        ldr_entry = self.readMemoryFormat(modlist, '<P')[0]
        while ldr_entry != modlist:
            ldte = self.getStruct('nt.LDR_DATA_TABLE_ENTRY', ldr_entry)
            dllname = self.readMemory(ldte.FullDllName.Buffer, ldte.FullDllName.Length).decode('utf-16le')
            dllbase = ldte.DllBase & self.bigmask
            self.addLibraryBase(dllname, dllbase, always=True)
            ldr_entry = ldte.InLoadOrderLinks.Flink & self.bigmask

        try:
            self.addBreakpoint(KeBugCheckBreak('nt.KeBugCheck'))
        except Exception as e:
            logger.warning('Error Seting KeBugCheck Bp: %s', e)

        try:
            self.addBreakpoint(KeBugCheckBreak('nt.KeBugCheckEx'))
        except Exception as e:
            logger.warning('Error Seting KeBugCheck Bp: %s', e)


GDB_BP_SOFTWARE     = 0
GDB_BP_HARDWARE     = 1
GDB_BP_WATCH_WRITE  = 2
GDB_BP_WATCH_READ   = 3
GDB_BP_WATCH_ACCESS = 4

class GdbStubTrace(
        vtrace.Trace,
        GdbStubMixin,
        v_base.TracerBase):

    def __init__(self, archname):

        # First things first, lets steal ourself an arch!
        envi.stealArchMethods(self, archname)
        vtrace.Trace.__init__(self, archname=archname)
        v_base.TracerBase.__init__(self)
        GdbStubMixin.__init__(self)

        self._break_after_bp = False    # We break *at* the bp

    # FIXME this should have a cleaner abstraction to allow for stuff...
    # platformActivateBreak / Watch!
    # platformDeactivateBreak / Watch!

    # FIXME we also need cleaner abstraction for checkBreakpoints
    # (some platforms stop *on* break and some stop *after...)

    #def _activateBreak(self, bp):
        ## For now, we don't support watchpoints...
        #if not bp.active:
            #addr = bp.resolveAddress(self)
            #self._cmdTransact('Z%d,%x,%x' % (GDB_BP_SOFTWARE,addr,1))

    #def _cleanupBreakpoints(self, force=False):
        #'''
        #Cleanup any non-fastbreak breakpoints.  This routine doesn't even get
        #called in the event of mode FastBreak=True.
        #'''
        #self.fb_bp_done = False
        #for bp in self.breakpoints.itervalues():
            ## No harm in calling deactivate on
            ## an inactive bp
            #if force or not bp.fastbreak:
                #self._cmdTransact('z%d,%x,%x' % (GDB_BP_SOFTWARE,bp.getAddress(),1))
                #bp.active = False
                ##bp.deactivate(self)

    #def _checkForBreak(self):
        #"""
        #Check to see if we've landed on a breakpoint, and if so
        #deactivate and step us past it.
#
        #WARNING: Unfortunatly, cause this is used immidiatly before
        #a call to run/wait, we must block briefly even for the GUI
        #"""
        ## Steal a reference because the step should
        ## clear curbp...
        #bp = self.curbp
        #if bp is not None and bp.isEnabled():
            ## We had to remove a check for active and a deactivate here...
            #orig = self.getMode("FastStep")
            #self.setMode("FastStep", True)
            #self.stepi()
            #self.setMode("FastStep", orig)
            #self._activateBreak(bp)

    def buildNewTrace(self):
        arch = self.getMeta('Architecture')
        newt = GdbStubTrace(arch)
        newt.setMeta('GdbServerHost', self.getMeta('GdbServerHost'))
        newt.setMeta('GdbServerPort', self.getMeta('GdbServerPort'))
        return newt




import unittest
import vivisect

import logging
import envi.common as e_common
import vivisect.tests.helpers as helpers

from vivisect.tests.switchdata import *

logger = logging.getLogger(__name__)
#e_common.initLogging(logger, 7)

class SwitchTest(unittest.TestCase):
    def test_amd64_ms_switch_0(self):
        base = 0x400000     # looks for reg at 0x3fb120
        fva = base + 0x4ee0
        vw = genMsSwitchWorkspace(MS_MAPS, base)
        vw.setMeta('DefaultCall', 'msx64call')
        vw.makeFunction(fva)
        funcblocks = vw.getFunctionBlocks(fva)
        funcblocks.sort()

        logger.debug("\n\n%s blocks:\n\t%r\n\t%r" % (self, funcblocks, cbs_ms_0))
        logger.debug("switches: %r" % vw.getVaSetRows('SwitchCases') )
        self.assertEqual(cbs_ms_0, vw.getFunctionBlocks(fva))
        self.assertEqual(5, len(vw.getXrefsFrom(base+0x11257)))

    def test_libc_switch_0(self):
        vw = genLinuxSwitchWorkspace(LIBC_MAPS, 0x500000)
        vw.makeFunction(0x500000)
        funcblocks = vw.getFunctionBlocks(0x500000)
        funcblocks.sort()

        logger.debug("\n\n%s blocks:\n\t%r\n\t%r" % (self, funcblocks, cbs_libc_0))
        logger.debug("switches: %r" % vw.getVaSetRows('SwitchCases') )
        self.assertEqual(cbs_libc_0, vw.getFunctionBlocks(0x500000))

    def test_amd64_walker_switch_0(self):
        vw = genLinuxSwitchWorkspace(WALKER_MAPS, 0x5ff3a0, arch='amd64')
        vw.makeFunction(0x600000)
        funcblocks = set(vw.getFunctionBlocks(0x600000))

        self.assertEqual(cbs_walker_0, funcblocks)

    def test_amd64_ls_switch(self):
        self.maxDiff = None
        vw = helpers.getTestWorkspace('linux', 'amd64', 'ls')

        '''
        .text:0x00404072  loc_00404072: [9 XREFS]
        .text:0x00404072  8d42ff           lea eax,dword [rdx - 1]
        .text:0x00404075  83f805           cmp eax,5
        .text:0x00404078  0f87d7e6ffff     ja loc_00402755
        .text:0x0040407e  ff24c5e8294100   jmp qword [ptr_sub_0040438d_004129e8 + rax * 8]
        '''
        '''
        .text:0x0040295a  e8a1f8ffff       call ls.plt_getopt_long    ;ls.plt_getopt_long()
        .text:0x0040295f  83f8ff           cmp eax,0xffffffff
        .text:0x00402962  0f8463060000     jz loc_00402fcb
        .text:0x00402968  0583000000       add eax,131
        .text:0x0040296d  3d12010000       cmp eax,274
        .text:0x00402972  760c             jbe switch_00402980

        .text:0x00402974  case_-83to-53_-51to-43_-3E_-39to-38_-36_-34to-33_-2Dto-2C_-2A_-28to-23_-1E_-19_-Ato-4_402974: [218 XREFS]
        .text:0x00402974  bf02000000       mov edi,2
        .text:0x00402979  e8626b0000       call sub_004094e0    ;sub_004094e0(2)
        .text:0x0040297e  6690             nop 
        .text:0x00402980  switch_00402980: [1 XREFS]
        .text:0x00402980  ff24c550214100   jmp qword [ptr_sub_00402be5_00412150 + rax * 8]    ;lower: 0x-83, upper: 0xc
        '''
        '''
        .text:0x00404874  loc_00404874: [1 XREFS]
        .text:0x00404874  410fb608         movzx ecx,byte [r8]
        .text:0x00404878  80f978           cmp cl,120
        .text:0x0040487b  7623             jbe loc_004048a0

        .text:0x004048a0  loc_004048a0: [1 XREFS]
        .text:0x004048a0  440fb6d9         movzx r11d,cl
        .text:0x004048a4  switch_004048a4: [0 XREFS]
        .text:0x004048a4  42ff24dd601d4100 jmp qword [ptr_case_0_40495d_00411d60 + r11 * 8]    ;lower: 0x0, upper: 0x78
        '''
        _do_test(self, vw, jmpva=0x402980, fva=0x402690, cbs=cbs_amd64_ls_402980, xrefcnt=57)
        _do_test(self, vw, jmpva=0x4048a4, fva=0x404740, cbs=cbs_amd64_ls_4048a4, xrefcnt=14)
        _do_test(self, vw, jmpva=0x40d57f, fva=0x40d840, cbs=cbs_amd64_ls_40d57f, xrefcnt=11)
        _do_test(self, vw, jmpva=0x410300, fva=0x4101a0, cbs=cbs_amd64_ls_410300, xrefcnt=10)
        # doesn't work correctly here:
        #_do_test(self, vw, jmpva=0x40407e, fva=0x404223, cbs=cbs_amd64_ls_40407e, xrefcnt=0)

    def test_amd64_chown_switch(self):
        self.maxDiff = None
        vw = helpers.getTestWorkspace('linux', 'amd64', 'chown')

        '''
        .text:0x020045b9  488d0d087e0000   lea rcx,qword [rip + $loc_0200c3c8]
        .text:0x020045c0  0fb6d3           movzx edx,bl
        .text:0x020045c3  48630491         movsxd rax,dword [rcx + rdx * 4]
        .text:0x020045c7  4801c8           add rax,rcx
        .text:0x020045ca  ffe0             jmp rax
        '''
        _do_test(self, vw, jmpva=0x20045ca, fva=0x02004050, cbs=cbs_amd64_chown_2004050, xrefcnt=17)

    def test_i386_static32_switch(self):
        self.maxDiff = None
        vw = helpers.getTestWorkspace('linux', 'i386', 'static32.llvm.elf')

        '''
        .text:0x0805eec0  81f980000000     cmp ecx,128
        .text:0x0805eec6  7318             jnc loc_0805eee0
        .text:0x0805eec8  e8b398feff       call static32_llvm.__weak___x86.get_pc_thunk.bx    ;static32_llvm.__weak___x86.get_pc_thunk.bx()
        .text:0x0805eecd  81c31beb0400     add ebx,0x0004eb1b
        .text:0x0805eed3  031c8b           add ebx,dword [ebx + ecx * 4]
        .text:0x0805eed6  01ca             add edx,ecx
        .text:0x0805eed8  switch_0805eed8: [0 XREFS]
        .text:0x0805eed8  ffe3             jmp ebx    ;lower: 0x20, upper: 0x7f
        '''
        _do_test(self, vw, jmpva=0x805eed8, fva=0x805edc0, cbs=cbs_i386_static_805edc0, xrefcnt=96)
        _do_test(self, vw, jmpva=0x80621da, fva=0x80621a0, cbs=cbs_i386_static_80621a0, xrefcnt=7)
        _do_test(self, vw, jmpva=0x808db0a, fva=0x808dc28, cbs=cbs_i386_static_808dc28, xrefcnt=9)
        _do_test(self, vw, jmpva=0x80a6954, fva=0x80a6920, cbs=cbs_i386_static_80a6920, xrefcnt=7)


    def test_i386_ld_switch(self):
        self.maxDiff = None
        vw = helpers.getTestWorkspace('linux', 'i386', 'ld-2.31.so')
        vw.analyze()

        cur_switches = [x for x, y, z in vw.getVaSetRows('SwitchCases')]
        cur_switches.sort()
        ld_switches.sort()
        for ld_switch in ld_switches:
            self.assertIn(ld_switch, cur_switches)

        '''
        .text:0x0200214c  loc_0200214c: [5 XREFS]
        .text:0x0200214c  8b55d0           mov edx,dword [ebp + local52]
        .text:0x0200214f  83ea06           sub edx,6
        .text:0x02002152  83fa23           cmp edx,35
        .text:0x02002155  7749             ja case_Eto13_15to28_2A_2Cto2E_20021a0
        .text:0x02002157  8b84970030ffff   mov eax,dword [edi + edx * 4 + -53248]
        .text:0x0200215e  01f8             add eax,edi
        .text:0x02002160  switch_02002160: [0 XREFS]
        .text:0x02002160  3effe0           ds: jmp eax    ;lower: 0xc, upper: 0x2f

        '''
        _do_test(self, vw, jmpva=0x2002160, fva=0x2001dd0, cbs=cbs_i386_ld_2002160, xrefcnt=6)
        #_do_test(self, vw, jmpva=0x80621da, fva=0x80621a0, cbs=cbs_i386_static_80621a0, xrefcnt=7)
        #_do_test(self, vw, jmpva=0x808db0a, fva=0x808dc28, cbs=cbs_i386_static_808dc28, xrefcnt=9)
        #_do_test(self, vw, jmpva=0x80a6954, fva=0x80a6920, cbs=cbs_i386_static_80a6920, xrefcnt=7)




def _do_test(self, vw, jmpva, fva, cbs, xrefcnt):
    vw.makeFunction(fva)
    funcblocks = vw.getFunctionBlocks(fva)
    funcblocks.sort()
    cbs.sort()
    logger.debug("\n\n%s (3) blocks:\n\t%r\n\t%r" % (self, funcblocks, cbs))
    logger.debug("switches: %r" % vw.getVaSetRows('SwitchCases') )
    self.assertEqual(cbs, funcblocks)
    self.assertEqual(xrefcnt, len(vw.getXrefsFrom(jmpva)))





'''
Perform further analysis on 64-bit GO language executables.

This module extends code flow analysis to runtime_main for Windows PE binaries.
GO binaries start from a single export and proceed thru several functions
that initialize GO.  Specific application code is launched from the GO function
runtime_main(), which is invoked by "call rax" with an address placed on the
stack many calls earlier.  Vivisect code flow analysis does not track the
address; this module finds the address and invokes makeFunction(va).

Samples have been found with different instruction sequences for reaching the
address of runtime_main(); this module attempts all observed sequences.
'''

import envi
import envi.archs.i386.disasm
import envi.archs.amd64.disasm
from vivisect.analysis.i386.golang import golang_collect_opcodes, find_golang_bblock

import logging

logger = logging.getLogger(__name__)

# Opcode sequence, where element [-6] is the important one.
_GOLANG_AMD64_INSTRS = ['cld', 'call', 'mov', 'mov', 'mov', 'mov', 'call',
                        'call', 'call', 'lea', 'push', 'push', 'call',
                        'pop', 'pop']

def analyze(vw):
    '''
    Perform further analysis on GO language executables.
    '''
    # Make sure it is a PE file, with "Go build ID:" in the first few bytes
    # of the .text segment (versus a .upxN segment of a packed sample).
    has_go_build = False
    for segment in vw.getSegments():
        va, size, name, filename = segment
        if name == '.text':
            bytez = vw.readMemory(va, 10000)
            k1 = bytez.find(b'Go build ID: ')
            if k1 != -1:
                has_go_build = True
                break
    if not has_go_build:
        return

    # Search the entry point (public export) for the pointer to runtime_main().
    # Most GO executables have a single entry point, but some are more complex.
    ep = vw.getEntryPoints()
    ptr_va, runtime_va = golang_search_eps(vw, ep, filename)
    if runtime_va is None:
        return

    # Invoke codeflow on runtime_main().
    vw.addEntryPoint(runtime_va)
    logger.debug('discovered runtime function: 0x%x', runtime_va)
    vw.makeFunction(runtime_va)

    # Also mark the ptr_va as a pointer to runtime_va.
    vw.makePointer(ptr_va, tova=runtime_va)


_GOLANG_AMD64_MEP1A_INSTRS \
    = ['sub', 'mov', 'mov', 'call', 'call', 'nop', 'nop', 'add', 'ret']
_GOLANG_AMD64_MEP1B_INSTRS \
    = ['sub', 'mov', 'call', 'call', 'nop', 'nop', 'add', 'ret']
_GOLANG_AMD64_MEP2A_INSTRS \
    = ['mov', 'mov', 'call', 'mov', 'mov', 'mov', 'mov', 'mov', 'mov', 'mov',
       'call', 'mov', 'mov', 'test', 'jz']
_GOLANG_AMD64_MEP2B_INSTRS \
    = ['mov', 'mov', 'call', 'mov', 'mov', 'mov', 'mov', 'mov', 'mov',
       'call', 'mov', 'mov', 'test', 'jz']
def golang_search_eps(vw, ep , filename):
    '''
    Search over all entry points to find main(), and then look for the
    function that calls runtime_main().
    Return two pointers, or None, None if not found.
    '''
    if len(ep) == 1:
        ep_va = ep[0]
        bblocks = vw.getFunctionBlocks(ep_va)
        if not bblocks:
            return None, None
        ptr_va, runtime_va = extract_golang_mainmain(vw, bblocks, filename)
        return ptr_va, runtime_va

    # Look for an entry point with a single basic block with a specific set
    # of instructions.  One of the call instructions has the next function.
    # There can be multiple candidate functions meeting this criteria.
    candidate_fns_1 = set()
    for next_ep in ep:
        bblocks = vw.getFunctionBlocks(next_ep)
        if (not bblocks) or (len(bblocks) != 1):
            continue
        instrs = golang_collect_opcodes(vw, bblocks[0])
        if (len(instrs) != len(_GOLANG_AMD64_MEP1A_INSTRS)) and \
           (len(instrs) != len(_GOLANG_AMD64_MEP1B_INSTRS)):
            continue
        instrs_mnem = [ op.mnem for op in instrs ]
        if (instrs_mnem != _GOLANG_AMD64_MEP1A_INSTRS) and \
           (instrs_mnem != _GOLANG_AMD64_MEP1B_INSTRS):
            continue
        opcode = instrs[-5]
        if not isinstance(opcode.opers[0],
                          envi.archs.i386.disasm.i386PcRelOper):
            continue
        try:
            candidate_fns_1.add(opcode.opers[0].getOperValue(opcode))
        except Exception:
            continue

    if not candidate_fns_1:
        return None, None

    # Next function has many basic blocks, one of which makes a call to main().
    candidate_fns_2 = set()
    for fnptr in candidate_fns_1:
        if not vw.isFunction(fnptr):
            continue
        bblocks = vw.getFunctionBlocks(fnptr)
        for bblock in bblocks:
            instrs = golang_collect_opcodes(vw, bblock)
            if (len(instrs) != len(_GOLANG_AMD64_MEP2A_INSTRS)) and \
               (len(instrs) != len(_GOLANG_AMD64_MEP2B_INSTRS)):
                continue
            instrs_mnem = [ op.mnem for op in instrs ]
            if (instrs_mnem != _GOLANG_AMD64_MEP2A_INSTRS) and \
               (instrs_mnem != _GOLANG_AMD64_MEP2B_INSTRS):
                continue
            opcode = instrs[-5]
            if not isinstance(opcode.opers[0],
                              envi.archs.i386.disasm.i386PcRelOper):
                continue
            try:
                candidate_fns_2.add(opcode.opers[0].getOperValue(opcode))
            except Exception:
                continue

    # There should be just one candidate function by now.
    if len(candidate_fns_2) != 1:
        return None, None

    ptr = candidate_fns_2.pop()

    # Analyze the function at the pointer if necessary.
    if not vw.isFunction(ptr):
        return None, None
    bblocks = vw.getFunctionBlocks(ptr)
    if not bblocks:
        return None, None

    # This might be the function that calls runtime_main(), or there might
    # be one more indirect jump in a single basic block.
    if len(bblocks) > 1:
        ptr_va, runtime_va = extract_golang_mainmain(vw, bblocks, filename)
        if runtime_va is None:
            return None, None
        return ptr_va, runtime_va

    # Look for the indirect jump.
    # Expect a function with one BB, with an indirect jump to the
    # address loaded by "lea rax,[rip + immediate]".
    instrs = golang_collect_opcodes(vw, bblocks[0])
    if not instrs:
        return None, None
    op = instrs[0]
    ptr_va, _ = parse_lea_raxriprel(vw, op, filename)
    if not vw.isFunction(ptr_va):
        return None, None
    bblocks = vw.getFunctionBlocks(ptr_va)
    ptr_va, runtime_va = extract_golang_mainmain(vw, bblocks, filename)
    return ptr_va, runtime_va


def extract_golang_mainmain(vw, basic_blocks, filename):
    '''
    Find the basic block of interest and return the address of
    the pointer and its contents, runtime_main().

    The BB will contain the sequence of opcodes in _GOLANG_AMD64_INSTRS.
    The push instruction -6 from the end will load the contents of a memory
    address, and those contents are the runtime_main() address.
    '''
    op = find_golang_bblock(vw, basic_blocks, _GOLANG_AMD64_INSTRS, 6)
    if op is None:
        op = find_golang_bblock_via_ind_jmp(vw, filename)
        if op is None:
            return None, None

    # The key opcode is "lea rax,[rip + immediate]", which points to
    # the GO function runtime_mainPC (aka runtime_main).
    ptr_va, runtime_va = parse_lea_raxriprel(vw, op, filename, get_content=True)
    return ptr_va, runtime_va


def find_golang_bblock_via_ind_jmp(vw, filename):
    '''
    Find the basic block of interest and return the address where
    the special sequence of opcodes begins.  Return None if not found.

    Some GO executables use an indirect jmp in the entry point,
    and special logic is needed to locate the BB of interest.
    Example:  1897d2de0837090ec350004ef2f9fa87.
    '''
    # Start from the entry point (we already know there is only one).
    ep_va = vw.getEntryPoints()[0]

    # Expect a function with one BB, ending with an indirect jump
    # to the address loaded by "lea rax,[rip + immediate]".
    basic_blocks = vw.getFunctionBlocks(ep_va)
    if len(basic_blocks) != 1:
        return None
    instrs = golang_collect_opcodes(vw, basic_blocks[0])
    if (len(instrs) < 2) or \
       (instrs[-1].mnem != 'jmp') or \
       (instrs[-2].mnem != 'lea'):
        return None
    op = instrs[-2]
    ptr_va, _ = parse_lea_raxriprel(vw, op, filename)
    if not ptr_va:
        return None

    # Analyze the function at the pointer if necessary.
    if not vw.isFunction(ptr_va):
        logger.debug('discovered new function (ptr): 0x%x', ptr_va)
        vw.makeFunction(ptr_va)

    # Expect a function with one BB, ending with an indirect jump
    # to the address loaded by "lea rax,[rip + immediate]".
    basic_blocks = vw.getFunctionBlocks(ptr_va)
    if len(basic_blocks) != 1:
        return None
    instrs = golang_collect_opcodes(vw, basic_blocks[0])
    if (len(instrs) < 2) or \
       (instrs[-1].mnem != 'jmp') or \
       (instrs[-2].mnem != 'lea'):
        return None
    op = instrs[-2]
    ptr_va, _ = parse_lea_raxriprel(vw, op, filename)
    if not ptr_va:
        return None

    # This function should contain the special basic block.
    basic_blocks = vw.getFunctionBlocks(ptr_va)
    return find_golang_bblock(vw, basic_blocks, _GOLANG_AMD64_INSTRS, 6)


def parse_lea_raxriprel(vw, opcode, filename, get_content=False):
    '''
    Parse an opcode that should be "lea rax,[rip + immediate]", returning
    the address of the second operand.  Also return the content of the
    address if so requested.
    Return None, None if there is an error.
    '''
    if len(opcode.opers) != 2:
        return None, None
    if not isinstance(opcode.opers[1], envi.archs.amd64.disasm.Amd64RipRelOper):
        return None, None
    try:
        ptr_va = opcode.opers[1].getOperValue(opcode)
        if len(vw.readMemory(ptr_va, 8)) != 8:
            return None, None
        if get_content:
            runtime_va = vw.castPointer(ptr_va)
            if len(vw.readMemory(runtime_va, 8)) != 8:
                return None, None
        else:
            runtime_va = None
        return ptr_va, runtime_va
    except Exception as e:
        return None, None



'''
Analysis plugin for supporting WorkspaceEmulators during analysis pass.
Finds and connects Switch Cases, most specifically from Microsoft.
'''

import envi
import envi.archs.i386 as e_i386
import envi.const as e_const

import vivisect.const as v_const
import vivisect.analysis.generic.codeblocks as vagc

import logging

logger = logging.getLogger(__name__)


def analyzeJmp(amod, emu, op, starteip):
    '''
    Top level logic
    '''
    vw = emu.vw
    ctx = getSwitchBase(vw, op, starteip, emu)
    if ctx is not None:
        tova, scale = ctx
        fva = vw.getFunction(starteip)
        vw.makePointer(tova, follow=False)
        vw.makeJumpTable(op, tova, rebase=True, psize=scale)
        # so the codeblocks this jumptable points to aren't proper locations...yet.
        # let's fix that up and kick off codeblock analysis to make the codeblocks
        for xrfrom, xrto, xrtype, xrflags in vw.getXrefsFrom(op.va, rtype=v_const.REF_CODE):
            vw.makeCode(xrto, fva=fva)
        vagc.analyzeFunction(vw, fva)


def getRealRegIdx(emu, regidx):
    return emu.getRegisterIndex(emu.getRealRegisterNameByIdx(regidx))


def findOp(vw, emu, startOp, mnem, regidx=None):
    cb = vw.getCodeBlock(startOp.va)
    backLoc = vw.getLocation(startOp.va - 1)
    while backLoc is not None and vw.getCodeBlock(backLoc[v_const.L_VA]) == cb:
        backOp = vw.parseOpcode(backLoc[0])
        if len(backOp.opers):
            oper = backOp.opers[0]
            if backOp.mnem == mnem:
                if regidx is not None:
                    if oper.isReg() and getRealRegIdx(emu, oper.reg) == regidx:
                        return backOp
                else:
                    return backOp
        backLoc = vw.getLocation(backLoc[0] - 1)


def scanUp(vw, emu, startva, regidx, valu):
    cb = vw.getCodeBlock(startva)
    loc = vw.getLocation(startva)
    while loc is not None and vw.getCodeBlock(loc[v_const.L_VA]) == cb:
        op = vw.parseOpcode(loc[0])
        if len(op.opers) > 1 and op.opers[0].isReg() and getRealRegIdx(emu, op.opers[0].reg) == regidx:
            if valu == emu.getOperValue(op, 1):
                return True
        loc = vw.getLocation(loc[0] - 1)

    return False


def getSwitchBase(vw, op, vajmp, emu=None):
    if not (op.iflags & envi.IF_BRANCH):
        return

    filename = vw.getMemoryMap(vajmp)[3]
    imgbase = vw.getFileMeta(filename, 'imagebase')

    if not op.opers[0].isReg():
        return

    reg = op.opers[0].reg
    if reg & e_const.RMETA_NMASK != reg:
        reg = getRealRegIdx(emu, reg)

    # Search up instructions until we get to the actual assignment of our
    # jump register, which should be an add in 64 bit town
    addOp = findOp(vw, emu, op, 'add', reg)
    if addOp is None:
        return

    regbase = addOp.getOperValue(1, emu)
    if regbase != imgbase:
        # just in case let's check a few more instructions up, because the first register could be
        # being used as the base instead (which means the second register is being used as the selector)
        if not scanUp(vw, emu, addOp.va, reg, imgbase):
            logger.info("0x%x: reg != imagebase (0x%x != 0x%x)" % (op.va, regbase, imgbase))
            return

    # Now find the instruction before the add that does the actual mov
    movOp = findOp(vw, emu, addOp, 'mov', reg)
    if movOp is None:
        # try the other one just in case
        reg = getRealRegIdx(emu, addOp.opers[1].reg)
        movOp = findOp(vw, emu, addOp, 'mov', reg)
        if movOp is None:
            return

    # TODO: Want a more arch-independent way of doing this
    arrayOper = movOp.opers[1]
    if not isinstance(arrayOper, e_i386.i386SibOper):
        logger.info("0x%x: arrayOper is not an i386SibOper: %s" % (op.va, repr(arrayOper)))
        return

    if arrayOper.scale % 4 != 0:
        logger.info("0x%x: arrayoper scale is wrong: (%d mod 4 != 0)" % (op.va, arrayOper.scale))
        return

    scale = arrayOper.scale
    disp = arrayOper.disp
    tova = disp + imgbase

    # now check for the byte array before that. this one is optional. first two are not.
    # but honestly, not a whole lot to do here other than make an xref
    indirOp = findOp(vw, emu, movOp, 'movzx', None)
    if indirOp is not None:
        if len(indirOp.opers):
            oper = indirOp.opers[1]
            if isinstance(oper, e_i386.i386SibOper) and oper.scale == 1:
                logger.info("0x%.8x (i:0x%.8x): Double deref (hitting a byte array offset into the offset-array)" % (vajmp, indirOp.va))
                indirVa = oper.disp + imgbase
                vw.addLocation(indirVa, 1, v_const.LOC_NUMBER, "DerefTable")
                vw.addXref(indirOp.va, indirVa, v_const.REF_DATA)

    return (tova, scale)


if 'vw' in globals():
    vw = globals()['vw']
    vw.vprint("Starting Switchcase Module...")
    for va, reprOp, flags in vw.getVaSetRows('DynamicBranches'):
        op = vw.parseOpcode(va)
        if op is None:
            vw.vprint("Cannot analyze none op at 0x%x" % va)
            continue
        analyzeJmp(None, vw.getEmulator(va=va), op, va)  # it doesn't use archmod anyway
    vw.vprint("Switchcase Done")



from envi.archs.aarch64.const import *
from envi.archs.aarch64 import sysregs

import envi.registers as e_reg

'''
Strategy:
    * Use only distinct register for Register Context (unique in each bank)
    * Store a lookup table for the different banks of registers, based on the 
        register data in proc_modes (see const.py)
    * Emulator does translation from register/mode to actual storage container
        using reg_table and some math (see _getRegIdx)
'''
META_B_BASE = 0x080000
META_H_BASE = 0x100000
META_W_BASE = 0x200000
META_D_BASE = 0x400000  # only for SIMD registers
META_Q_BASE = 0

meta_reg_bases = (0,
        META_B_BASE,
        META_H_BASE,
        0,
        META_W_BASE,
        0,
        0,
        0,
        META_D_BASE,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        META_Q_BASE,
)

aarch64_regs_tups = [('x%d'%x, 64) for x in range(31)]  # x31 is zero register

REG_SP = len(aarch64_regs_tups)
aarch64_regs_tups.append(('sp', 64))
REG_PC = len(aarch64_regs_tups)
aarch64_regs_tups.append(('pc', 64))
REG_CPSR = len(aarch64_regs_tups)
aarch64_regs_tups.append(('cpsr', 64))
REG_XZR = len(aarch64_regs_tups)
aarch64_regs_tups.append(('xzr', 64))

REG_PSTATE = REG_CPSR

MAX_REGS = len(aarch64_regs_tups) - 1

# metas:   zr is x31.  w* is 32-bit versions of x* regs

reg_data = aarch64_regs_tups
aarch64_regs = [r for r,sz in aarch64_regs_tups]

# Defining general registers 0 - 30
aarch64_metas = [("w%d" % x, x, 0, 32) for x in range(31)]          
aarch64_metas.extend([("h%d" % x, x, 0, 16) for x in range(31)])
aarch64_metas.extend([("b%d" % x, x, 0, 8) for x in range(31)])

# Adding register 31 context (wzr or xzr) separately
aarch64_metas.append(("wzr", REG_XZR, 0, 32))

REG_FPSCR = len(reg_data)
reg_data.append(('fpcr', 32))
reg_data.append(('fpsr', 32))

REGS_SYSTEM_REGS = len(reg_data)

# System Registers - build from sysregs module
sys_regs = [(reg, 64) for reg, _, _, _, _, _, _ in sysregs.system_registers]
reg_data.extend(sys_regs)

# VFP Registers
#    VFPv2 consists of 16 doubleword registers
#    VFPv3 either 32 or 16 doubleword registers
#       VFPv3-D32 - implementation with 32 doubleword regs
#       VFPv3-D16 - implementation with 16 doubleword regs
#    VFPv4 - same as VFPv3
# Advanced SIMD and Floating point views are *not* identical
#
# we implement VFPv4-D32 since it should be backwards-compatible with all others
#  the largest accessor of this extended register bank is 128bits so we'll go with that.

REGS_VECTOR_BASE_IDX = len(reg_data)
for simdreg in range(VFP_QWORD_REG_COUNT):
    simd_idx = REGS_VECTOR_BASE_IDX + simdreg
    d = simdreg #* 2
    s = d #* 2
    reg_data.append(("v%d" % simdreg, 256))
    aarch64_metas.append(("q%d" % (d),   simd_idx, 0, 128))
    aarch64_metas.append(("d%d" % (d),   simd_idx, 0, 64))
    aarch64_metas.append(("s%d" % (s),   simd_idx, 0, 32))
    aarch64_metas.append(("h%d" % (s),   simd_idx, 0, 16))
    aarch64_metas.append(("b%d" % (s),   simd_idx, 0, 8))

l = locals()
e_reg.addLocalEnums(l, aarch64_regs_tups)
REG_LR = REG_X30



PSR_N = 31  # negative
PSR_Z = 30  # zero
PSR_C = 29  # carry
PSR_V = 28  # oVerflow
PSR_Q = 27
PSR_IT = 25
PSR_J = 24
PSR_DNM = 20
PSR_GE = 16
PSR_E = 9
PSR_A = 8
PSR_I = 7
PSR_F = 6
PSR_T = 5
PSR_M = 0

PSR_N_bit  = 1 << PSR_N
PSR_N_mask = 0xffffffff ^ PSR_N_bit
PSR_Z_bit  = 1 << PSR_Z
PSR_Z_mask = 0xffffffff ^ PSR_Z_bit
PSR_C_bit  = 1 << PSR_C
PSR_C_mask = 0xffffffff ^ PSR_C_bit
PSR_V_bit  = 1 << PSR_V
PSR_V_mask = 0xffffffff ^ PSR_V_bit
PSR_T_bit  = 1 << PSR_T
PSR_T_mask = 0xffffffff ^ PSR_T_bit

psr_fields = [None for x in range(32)]
psr_fields[PSR_M] = "M"
psr_fields[PSR_T] = "T"
psr_fields[PSR_F] = "F"
psr_fields[PSR_I] = "I"
psr_fields[PSR_A] = "A"
psr_fields[PSR_E] = "E"
psr_fields[PSR_GE] = "GE"
psr_fields[PSR_GE+1] = "GE+1"
psr_fields[PSR_GE+2] = "GE+2"
psr_fields[PSR_GE+3] = "GE+3"
psr_fields[PSR_DNM] = "DNM"
psr_fields[PSR_DNM+1] = "DNM+1"
psr_fields[PSR_DNM+2] = "DNM+2"
psr_fields[PSR_DNM+3] = "DNM+3"
psr_fields[PSR_J] = "J"
psr_fields[PSR_IT] = "IT"
psr_fields[PSR_IT+1] = "IT+1"
psr_fields[PSR_IT-15] = "IT+2"  # IT is split into two sections
psr_fields[PSR_IT-14] = "IT+3"
psr_fields[PSR_IT-13] = "IT+4"
psr_fields[PSR_IT-12] = "IT+5"
psr_fields[PSR_IT-11] = "IT+6"
psr_fields[PSR_IT-10] = "IT+7"
psr_fields[PSR_Q] = "Q"
psr_fields[PSR_V] = "V"
psr_fields[PSR_C] = "C"
psr_fields[PSR_Z] = "Z"
psr_fields[PSR_N] = "N"

aarch64_status_metas = [
        ("N", REG_PSTATE, PSR_N, 1, "Negative/LessThan flag"),
        ("Z", REG_PSTATE, PSR_Z, 1, "Zero flag"),
        ("C", REG_PSTATE, PSR_C, 1, "Carry/Borrow/Extend flag"),
        ("V", REG_PSTATE, PSR_V, 1, "oVerflow flag"),
        ("Q", REG_PSTATE, PSR_Q, 1, "Sticky Overflow flag"),
        ("J", REG_PSTATE, PSR_J, 1, "Jazelle Mode bit"),
        ("GE",REG_PSTATE, PSR_GE, 4, "Greater/Equal flag"),
        ("DNM",REG_PSTATE, PSR_DNM, 4, "DO NOT MODIFY bits"),
        ("IT0",REG_PSTATE, PSR_IT, 1, "IfThen 0 bit"),
        ("IT1",REG_PSTATE, PSR_IT+1, 1, "IfThen 1 bit"),
        ("IT2",REG_PSTATE, PSR_IT+2, 1, "IfThen 2 bit"),
        ("IT3",REG_PSTATE, PSR_IT+3, 1, "IfThen 3 bit"),
        ("IT4",REG_PSTATE, PSR_IT+4, 1, "IfThen 4 bit"),
        ("IT5",REG_PSTATE, PSR_IT+5, 1, "IfThen 5 bit"),
        ("IT6",REG_PSTATE, PSR_IT+6, 1, "IfThen 6 bit"),
        ("IT7",REG_PSTATE, PSR_IT+7, 1, "IfThen 7 bit"),
        ("E", REG_PSTATE, PSR_E, 1, "Data Endian bit"),
        ("A", REG_PSTATE, PSR_A, 1, "Imprecise Abort Disable bit"),
        ("I", REG_PSTATE, PSR_I, 1, "IRQ disable bit"),
        ("F", REG_PSTATE, PSR_F, 1, "FIQ disable bit"),
        ("T", REG_PSTATE, PSR_T, 1, "Thumb Mode bit"),
        ("M", REG_PSTATE, PSR_M, 5, "Processor Mode"),
        ]

e_reg.addLocalStatusMetas(l, aarch64_metas, aarch64_status_metas, "CPSC")
e_reg.addLocalMetas(l, aarch64_metas)


class A64RegisterContext(e_reg.RegisterContext):
    def __init__(self):
        e_reg.RegisterContext.__init__(self)
        self.loadRegDef(reg_data)
        self.loadRegMetas(aarch64_metas, statmetas=aarch64_status_metas)
        self.setRegisterIndexes(REG_PC, REG_SP)

rctx = A64RegisterContext()




'''
A quick set of tools for doing stack doping.
'''
import logging

import vtrace

logger = logging.getLogger(__name__)


def dopeThreadStack(trace, threadid):
    curthread = trace.getCurrentThread()
    try:
        trace.selectThread(threadid)
        sp = trace.getStackCounter()
        mmap = trace.getMemoryMap(sp)
        if mmap is None:
            raise Exception('Thread %d has invalid stack pointer 0x%.8x' % (threadid, sp))

        mapva, mapsize, mperms, mfname = mmap

        dopesize = sp - mapva
        trace.writeMemory(mapva, 'V' * dopesize)

    except Exception as e:
        logger.warning('dopeThreadStack Failed On %d (%s)', threadid, e)
        trace.selectThread(curthread)


def dopeAllThreadStacks(trace):
    '''
    Apply stack doping to all thread stacks.
    '''
    for threadid in trace.getThreads().keys():
        dopeThreadStack(trace, threadid)


class ThreadDopeNotifier(vtrace.Notifier):

    def notify(self, event, trace):
        dopeAllThreadStacks(trace)


dopenotif = ThreadDopeNotifier()


def enableEventDoping(trace):
    trace.registerNotifier(vtrace.NOTIFY_CONTINUE, dopenotif)


def disableEventDoping(trace):
    trace.deregisterNotifier(vtrace.NOTIFY_CONTINUE, dopenotif)



import binascii
import unittest

import envi
import envi.memcanvas as e_memcanvas

import vivisect

# name, bytes, va, repr, txtRender
i386SingleByteOpcodes = [
    ('add', '0001', 0x40, 'add byte [ecx],al', 'add byte [ecx],al'),
    ('adc', '83D306', 0x40, 'adc ebx,6', 'adc ebx,6'),
    ('jg', '7faa', 0x400, 'jg 0x000003ac', 'jg 0x000003ac'),
    ('rep movsb', 'f3a4', 0x40, 'rep: movsb ', 'rep: movsb '),
    ('mov al', 'b0aa', 0x40, 'mov al,170', 'mov al,170'),
    ('mov ebx', 'b8aaaa4040', 0x40, 'mov eax,0x4040aaaa', 'mov eax,0x4040aaaa'),
    ('inc ebx', '43', 0x40, 'inc ebx', 'inc ebx'),
    ('call ebx', 'ffd3', 0x40, 'call ebx', 'call ebx'),
    ('call lit', 'e801010101', 0x40, 'call 0x01010146', 'call 0x01010146'),
    ('mov dword', '89aa41414141', 0x40, 'mov dword [edx + 1094795585],ebp', 'mov dword [edx + 1094795585],ebp'),
    ('imul 1', 'f6aaaaaaaaaa', 0x40, 'imul byte [edx - 1431655766]', 'imul byte [edx - 1431655766]'),
    ('imul 2', 'f7aaaaaaaaaa', 0x40, 'imul dword [edx - 1431655766]', 'imul dword [edx - 1431655766]'),
    ('push', 'fff0', 0x40, 'push eax', 'push eax'),
    ('push 2', '6aff', 0x40, 'push 0xffffffff', 'push 0xffffffff'),
    ('push 3', '68ffffffff', 0x40, 'push 0xffffffff', 'push 0xffffffff'),
    ('nop 1', '90', 0x40, 'nop ', 'nop '),
    ('nop 2', '0f1fc0', 0x40, 'nop eax', 'nop eax'),
    ('nop 3', 'f30f1fc0', 0x40, 'rep: nop eax', 'rep: nop eax'),
    ('pop', '8ff0', 0x40, 'pop eax', 'pop eax'),
    ('pop', '8ffb', 0x40, 'pop ebx', 'pop ebx'),
    ('BSWAP (eax)', '0fc8', 0x40, 'bswap eax', 'bswap eax'),
    ('BSWAP (ebx)', '0fcb', 0x40, 'bswap ebx', 'bswap ebx'),
    ('BSWAP (eax)', 'f30fc8', 0x40, 'rep: bswap eax', 'rep: bswap eax'),
    ('BSWAP (ebx)', 'f30fcb', 0x40, 'rep: bswap ebx', 'rep: bswap ebx'),
    ('setg (al)', '0f9fc0', 0x40, 'setg al', 'setg al'),
    ('setg (dl)', '0f9fc2', 0x40, 'setg dl', 'setg dl'),
    ('rep setg (al)', 'f30f9fc0', 0x40, 'rep: setg al', 'rep: setg al'),
    ('rep setg (dl)', 'f30f9fc2', 0x40, 'rep: setg dl', 'rep: setg dl'),
    ('prefix scas', 'f2ae', 0x40, 'repnz: scasb ', 'repnz: scasb '),
    ('jmp', 'E910000000', 0x40, 'jmp 0x00000055', 'jmp 0x00000055'),
    ('TEST', '84db', 0x40, 'test bl,bl', 'test bl,bl'),
    ('POP', '5D', 0x40, 'pop ebp', 'pop ebp'),
    ('POP', '5B', 0x40, 'pop ebx', 'pop ebx'),
    ('nop', '660f1cf8', 0x40, 'nop ax', 'nop ax'),
    ('nop', '660f1df8', 0x40, 'nop ax', 'nop ax'),
    ('nop', '660f1ef8', 0x40, 'nop ax', 'nop ax'),
    ('nop', '660f1ff8', 0x40, 'nop ax', 'nop ax'),
    ('nop', 'f20f1cfb', 0x40, 'repnz: nop ebx', 'repnz: nop ebx'),
    ('nop', 'f20f1dfb', 0x40, 'repnz: nop ebx', 'repnz: nop ebx'),
    ('nop', 'f20f1efb', 0x40, 'repnz: nop ebx', 'repnz: nop ebx'),
    ('nop', 'f20f1ffb', 0x40, 'repnz: nop ebx', 'repnz: nop ebx'),
    ('nop', 'f30f1cfa', 0x40, 'rep: nop edx', 'rep: nop edx'),
    ('nop', 'f30f1dfa', 0x40, 'rep: nop edx', 'rep: nop edx'),
    ('nop', 'f30f1efa', 0x40, 'rep: nop edx', 'rep: nop edx'),
    ('nop', 'f30f1ffa', 0x40, 'rep: nop edx', 'rep: nop edx'),
    ('nop', '0f1cfa', 0x40, 'nop edx', 'nop edx'),
    ('nop', '0f1dfa', 0x40, 'nop edx', 'nop edx'),
    ('nop', '0f1efa', 0x40, 'nop edx', 'nop edx'),
    ('nop', '0f1ffa', 0x40, 'nop edx', 'nop edx'),
    ('DIV 1', '66f7f2', 0x40, 'div dx', 'div dx'),
    ('DIV 2', 'f7f1', 0x40, 'div ecx', 'div ecx'),
    ('RCR 1', '66d1d8', 0x40, 'rcr ax,1', 'rcr ax,1'),
    ('ROL', 'd3c3', 0x40, 'rol ebx,cl', 'rol ebx,cl'),
]

i386MultiByteOpcodes = [
    ('CMPXCH8B', 'f00fc74d00', 0x40, 'lock: cmpxch8b qword [ebp]', 'lock: cmpxch8b qword [ebp]'),
    ('jmp 2', 'FF248D3A3A3A3A', 0x40, 'jmp dword [0x3a3a3a3a + ecx * 4]', 'jmp dword [0x3a3a3a3a + ecx * 4]'),
    ('MOV', '8B148541414141', 0x40, 'mov edx,dword [0x41414141 + eax * 4]', 'mov edx,dword [0x41414141 + eax * 4]'),
    ('MOV 2r', '678B14', 0x40, 'mov edx,dword [si]', 'mov edx,dword [si]'),
    ('PUSH 2', '67FF37', 0x40, 'push dword [bx]', 'push dword [bx]'),
    ('PUSH 3', '67ff30', 0x40, 'push dword [bx + si]', 'push dword [bx + si]'),
    ('PUSH 4', '67ff7020', 0x40, 'push dword [bx + si + 32]', 'push dword [bx + si + 32]'),
    ('ROR', '66d3c8', 0x40, 'ror ax,cl', 'ror ax,cl'),
    ('PEXTRB', '660F3A14D011', 0x40, 'pextrb eax,xmm2,17', 'pextrb eax,xmm2,17'),
    ('PEXTRB 2', '660F3A141011', 0x40, 'pextrb byte [eax],xmm2,17', 'pextrb byte [eax],xmm2,17'),
    ('MOV 3', '8B16', 0x40, 'mov edx,dword [esi]', 'mov edx,dword [esi]'),
    ('MOV 4', '66b90202', 0x40, 'mov cx,514', 'mov cx,514'),
    ('NOT', '66F7D0', 0x40, 'not ax', 'not ax'),
    ('NOT 2', 'F7D0', 0x40, 'not eax', 'not eax'),
    ('PUSH', '6653', 0x40, 'push bx', 'push bx'),
    ('CVTPD2PI', '660f2Daaaaaaaa41', 0x40, 'cvtpd2pi mm5,oword [edx + 1101703850]', 'cvtpd2pi mm5,oword [edx + 1101703850]'),
    ('CVTTPS2PI', '0f2caaaaaaaa41', 0x40, 'cvttps2pi mm5,qword [edx + 1101703850]', 'cvttps2pi mm5,qword [edx + 1101703850]'),
    ('CVTTSS2SI', 'f30f2caaaaaaaa41', 0x40, 'cvttss2si ebp,dword [edx + 1101703850]', 'cvttss2si ebp,dword [edx + 1101703850]'),
    ('CVTTSS2SI 2', 'f30f2ce9', 0x40, 'cvttss2si ebp,xmm1', 'cvttss2si ebp,xmm1'),
    ('CVTTPD2PI', '660f2caaaaaaaa41', 0x40, 'cvttpd2pi mm5,oword [edx + 1101703850]', 'cvttpd2pi mm5,oword [edx + 1101703850]'),
    ('CVTTSD2SI 2', 'f20f2caaaaaaaa41', 0x40, 'cvttsd2si ebp,qword [edx + 1101703850]', 'cvttsd2si ebp,qword [edx + 1101703850]'),
    ('ADDPS', '0f58aa41414141', 0x40, 'addps xmm5,oword [edx + 1094795585]', 'addps xmm5,oword [edx + 1094795585]'),
    ('MOVAPS', '0f28aa41414141', 0x40, 'movaps xmm5,oword [edx + 1094795585]', 'movaps xmm5,oword [edx + 1094795585]'),
    ('MOVAPD', '660f28aa41414141', 0x40, 'movapd xmm5,oword [edx + 1094795585]', 'movapd xmm5,oword [edx + 1094795585]'),
    ('RSM', '0faa', 0x40, 'rsm ', 'rsm '),
    ('MUL', 'f7e3', 0x40, 'mul ebx', 'mul ebx'),
    ('MUL 2', '66f7e3', 0x40, 'mul bx', 'mul bx'),
    ('PMULLW (66)', '660fd5cd', 0x40, 'pmullw xmm1,xmm5', 'pmullw xmm1,xmm5'),
    ('PMULLW', '0fd5e2', 0x40, 'pmullw mm4,mm2', 'pmullw mm4,mm2'),
    ('CMPXCH8B', '0fc70a', 0x40, 'cmpxch8b qword [edx]', 'cmpxch8b qword [edx]'),
    ('MOVD (66)', '660f7ecb', 0x40, 'movd ebx,xmm1', 'movd ebx,xmm1'),
    ('MOVD', '0F6E0D41414100', 0x40, 'movd mm1,dword [0x00414141]', 'movd mm1,dword [0x00414141]'),
    ('MOVQ', '0F6FCB', 0x40, 'movq mm1,mm3', 'movq mm1,mm3'),
    ('PSRAW',  '0FE1CA', 0x40, 'psraw mm1,mm2', 'psraw mm1,mm2'),
    ('PSLLQ (66)',  '660FF3CB', 0x40, 'psllq xmm1,xmm3', 'psllq xmm1,xmm3'),
    ('PALIGNR', '0F3A0FDC03', 0x40, 'palignr xmm3,xmm4,3', 'palignr xmm3,xmm4,3'),
    ('PALIGNR (66)',  '660F3A0FCA07', 0x40, 'palignr xmm1,xmm2,7', 'palignr xmm1,xmm2,7'),
    ('PSLLQ (reg)',  '660FF3CA', 0x40, 'psllq xmm1,xmm2', 'psllq xmm1,xmm2'),
    ('PSLLW (regs)',  '0F71F108', 0x40, 'psllw mm1,8', 'psllw mm1,8'),
    ('PSLLQ (66) 2',  '660F73F108', 0x40, 'psllq xmm1,8', 'psllq xmm1,8'),
    ('PSRLW (66)',  '660F71D611', 0x40, 'psrlw xmm6,17', 'psrlw xmm6,17'),
    ('PSRAD (66)',  '660F72E704', 0x40, 'psrad xmm7,4', 'psrad xmm7,4'),
    ('PSRLQ (66)',  '660F73D308', 0x40, 'psrlq xmm3,8', 'psrlq xmm3,8'),
    ('PSRAW (66)',  '660F71E108', 0x40, 'psraw xmm1,8', 'psraw xmm1,8'),
    ('PSRLDQ (66)', '660f73faaa', 0x40, 'pslldq xmm2,170', 'pslldq xmm2,170'),
    ('LFENCE', '0faee8', 0x40, 'lfence ', 'lfence '),
    ('LFENCE', '0faeea', 0x40, 'lfence ', 'lfence '),
    ('LFENCE', '0faeec', 0x40, 'lfence ', 'lfence '),
    ('LFENCE', '0faeef', 0x40, 'lfence ', 'lfence '),
    ('MFENCE', '0faef0', 0x40, 'mfence ', 'mfence '),
    ('MFENCE', '0faef2', 0x40, 'mfence ', 'mfence '),
    ('MFENCE', '0faef5', 0x40, 'mfence ', 'mfence '),
    ('MFENCE', '0faef7', 0x40, 'mfence ', 'mfence '),
    ('CVTTSS2SI', 'f30f2cc8', 0x40, 'cvttss2si ecx,xmm0', 'cvttss2si ecx,xmm0'),
    ('PREFIXES', 'f0673e2e65f30f10ca', 0x40, 'lock cs ds gs: movss xmm1,xmm2', 'lock cs ds gs: movss xmm1,xmm2'),
    ('BSR', '0fbdc3', 0x40, 'bsr eax,ebx', 'bsr eax,ebx'),
    ('LZNT', 'f30fbdc3', 0x40, 'lzcnt eax,ebx', 'lzcnt eax,ebx'),
    # Because of how the MODRM Bytes are set, these map to the same instruction
    # TODO: Would these be the same to a real x86 chip?
    ('PSRLDQ (66)', '660f73b5aa', 0x40, 'psllq xmm5,170', 'psllq xmm5,170'),
    ('PSRLDQ (66)', '660f73f5aa', 0x40, 'psllq xmm5,170', 'psllq xmm5,170'),
    # Same for these two
    ('PSRLDQ (66)', '660f73b1aa', 0x40, 'psllq xmm1,170', 'psllq xmm1,170'),
    ('PSRLDQ (66)', '660f73b9aa', 0x40, 'pslldq xmm1,170', 'pslldq xmm1,170'),

    ('ADDSS', 'f30f58ca', 0x40, 'addss xmm1,xmm2', 'addss xmm1,xmm2'),
    ('ADDSS 2', 'f30f580a', 0x40, 'addss xmm1,dword [edx]', 'addss xmm1,dword [edx]'),
    ('ADDSS 3', 'f30f585963', 0x40, 'addss xmm3,dword [ecx + 99]', 'addss xmm3,dword [ecx + 99]'),
    ('SQRTSS', 'f30f51d7', 0x40, 'sqrtss xmm2,xmm7', 'sqrtss xmm2,xmm7'),
    ('SQRTSS 2', 'f30f511c2541414141', 0x40, 'sqrtss xmm3,dword [0x41414141]', 'sqrtss xmm3,dword [0x41414141]'),
    ('SQRTSS 3', 'f30f511cd540000000', 0x40, 'sqrtss xmm3,dword [0x00000040 + edx * 8]', 'sqrtss xmm3,dword [0x00000040 + edx * 8]'),
    ('RSQRTSS', 'f30f52d7', 0x40, 'rsqrtss xmm2,xmm7', 'rsqrtss xmm2,xmm7'),
    ('RSQRTSS 2', 'f30f521c2541414141', 0x40, 'rsqrtss xmm3,dword [0x41414141]', 'rsqrtss xmm3,dword [0x41414141]'),
    ('RSQRTSS 3', 'f30f521cd540000000', 0x40, 'rsqrtss xmm3,dword [0x00000040 + edx * 8]', 'rsqrtss xmm3,dword [0x00000040 + edx * 8]'),
    ('RCPSS', 'f30f53cf', 0x40, 'rcpss xmm1,xmm7', 'rcpss xmm1,xmm7'),
    ('RCPSS 2', 'f30f5319', 0x40, 'rcpss xmm3,dword [ecx]', 'rcpss xmm3,dword [ecx]'),

    ('PCMPISTRI', '660f3a630f0d', 0x40, 'pcmpistri xmm1,oword [edi],13', 'pcmpistri xmm1,oword [edi],13'),
    ('PSHUFB', '660F3800EF', 0x40, 'pshufb xmm5,xmm7', 'pshufb xmm5,xmm7'),
    ('PSHUFB 2', '0F3800CA', 0x40, 'pshufb mm1,mm2', 'pshufb mm1,mm2'),
    ('PABSB', '660F381C08', 0x40, 'pabsb xmm1,oword [eax]', 'pabsb xmm1,oword [eax]'),
    ('PABSB 2', '0F381C1D41414141', 0x40, 'pabsb mm3,qword [0x41414141]', 'pabsb mm3,qword [0x41414141]'),
    ('RDTSC', '0F31', 0x40, 'rdtsc ', 'rdtsc '),
    ('RDTSCP', '0F01F9', 0x40, 'rdtscp ', 'rdtscp '),
    ('CVTDQ2PD', 'f30fe6c0', 0x40, 'cvtdq2pd xmm0,xmm0', 'cvtdq2pd xmm0,xmm0'),
    ('MOVQ', 'F30F7ECB', 0x40, 'movq xmm1,xmm3', 'movq xmm1,xmm3'),
    ('MOVQ (F3)', 'F30F7E0D41414100', 0x40, 'movq xmm1,qword [0x00414141]', 'movq xmm1,qword [0x00414141]'),
    ('MOVSD', 'f20f10ca', 0x40, 'movsd xmm1,xmm2', 'movsd xmm1,xmm2'),
    ('MOVSD (PREFIX)', 'f3f20f10ca', 0x40, 'rep: movsd xmm1,xmm2', 'rep: movsd xmm1,xmm2'),
    ('MOVSS (PREFIX)', 'f2f30f10ca', 0x40, 'repnz: movss xmm1,xmm2', 'repnz: movss xmm1,xmm2'),
    ('POPCNT', '66f30fb8c3', 0x40, 'popcnt ax,bx', 'popcnt ax,bx'),
    ('POPCNT', 'f30fb8c4', 0x40, 'popcnt eax,esp', 'popcnt eax,esp'),
    ('POPCNT', 'f30fb80541414100', 0x40, 'popcnt eax,dword [0x00414141]', 'popcnt eax,dword [0x00414141]'),
    ('LZCNT', 'f30fbdc4', 0x40, 'lzcnt eax,esp', 'lzcnt eax,esp'),
    ('LZCNT', 'f30fbd0541414100', 0x40, 'lzcnt eax,dword [0x00414141]', 'lzcnt eax,dword [0x00414141]'),
    ('MOVDQU', 'F30F6FCA', 0x40, 'movdqu xmm1,xmm2', 'movdqu xmm1,xmm2'),
    ('MOVDQU (MEM)', 'F30F6F4810', 0x40, 'movdqu xmm1,oword [eax + 16]', 'movdqu xmm1,oword [eax + 16]'),
    ('MOVDQU (REP)', 'F3F30F6FCA', 0x40, 'movdqu xmm1,xmm2', 'movdqu xmm1,xmm2'),
    ('MOVSD', 'f20f100d28330608', 0x40, 'movsd xmm1,qword [0x08063328]', 'movsd xmm1,qword [0x08063328]'),
    ('MOVSD 2', 'f20f1145f0', 0x40, 'movsd qword [ebp - 16],xmm0', 'movsd qword [ebp - 16],xmm0'),
    ('MOVSD 3', 'f20f100d70790908', 0x40, 'movsd xmm1,qword [0x08097970]', 'movsd xmm1,qword [0x08097970]'),
    ('MOVSS', 'f30f1045f8', 0x40, 'movss xmm0,dword [ebp - 8]', 'movss xmm0,dword [ebp - 8]'),
    ('MOVSS 2', 'f30f1055d0', 0x40, 'movss xmm2,dword [ebp - 48]', 'movss xmm2,dword [ebp - 48]'),
    ('MOVSS 3', 'F30F110D41414100', 0x40, 'movss dword [0x00414141],xmm1', 'movss dword [0x00414141],xmm1'),
    ('CVTSI2SS', 'f30f2ac8', 0x40, 'cvtsi2ss xmm1,eax', 'cvtsi2ss xmm1,eax'),
    ('MULSS', 'f30f59ca', 0x40, 'mulss xmm1,xmm2', 'mulss xmm1,xmm2'),
    ('SUBSS', 'f30f5cc1', 0x40, 'subss xmm0,xmm1', 'subss xmm0,xmm1'),
    ('CVTSS2SD', 'f30f5ac0', 0x40, 'cvtss2sd xmm0,xmm0', 'cvtss2sd xmm0,xmm0'),
    ('SQRTSS', 'f30f51d9', 0x40, 'sqrtss xmm3,xmm1', 'sqrtss xmm3,xmm1'),
    ('VMXON', 'f30fc73541414100', 0x40, 'vmxon qword [0x00414141]', 'vmxon qword [0x00414141]'),
    ('MULSD', 'f20f59c1', 0x40, 'mulsd xmm0,xmm1', 'mulsd xmm0,xmm1'),
    ('VMCLEAR', '660FC73541414100', 0x40, 'vmclear qword [0x00414141]', 'vmclear qword [0x00414141]'),
    ('CRC 1', 'f20f38f0e8', 0x40, 'crc32 ebp,al', 'crc32 ebp,al'),
    ('CRC 2', '66f20f38f1C3', 0x40, 'crc32 eax,bx', 'crc32 eax,bx'),
    ('CRC 3', 'f20f38f1C3', 0x40, 'crc32 eax,ebx', 'crc32 eax,ebx'),
    ('CLAC', '0f01ca', 0x40, 'clac ', 'clac '),
    ('STAC', '0f01cb', 0x40, 'stac ', 'stac '),
    ('VMFUNC', '0f01d4', 0x40, 'vmfunc ', 'vmfunc '),
    ('XEND', '0f01d5', 0x40, 'xend ', 'xend '),
    ('XGETBV', '0f01d0', 0x40, 'xgetbv ecx', 'xgetbv ecx'),
    ('XSETBV', '0f01d1', 0x40, 'xsetbv ecx', 'xsetbv ecx'),
    ('XADD', '660fc1d8', 0x40, 'xadd ax,bx', 'xadd ax,bx'),
    ('XADD 2', '0fc1d8', 0x40, 'xadd eax,ebx', 'xadd eax,ebx'),
    ('XTEST', '0f01d6', 0x40, 'xtest ', 'xtest '),
    ('MOVUPD', '660f10cc', 0x40, 'movupd xmm1,xmm4', 'movupd xmm1,xmm4'),
    ('MOVUPD', '660f1018', 0x40, 'movupd xmm3,oword [eax]', 'movupd xmm3,oword [eax]'),
    ('UNPCKLPD', '660F14A241414100', 0x40, 'unpcklpd xmm4,oword [edx + 4276545]', 'unpcklpd xmm4,oword [edx + 4276545]'),
    ('MOVHPD', '660F162541414141', 0x40, 'movhpd xmm4,oword [0x41414141]', 'movhpd xmm4,oword [0x41414141]'),
    ('MOVHPD', '660F172D41414141', 0x40, 'movhpd oword [0x41414141],xmm5', 'movhpd oword [0x41414141],xmm5'),
    ('MOVMSKPS', '0F50D6', 0x40, 'movmskps edx,xmm6', 'movmskps edx,xmm6'),
    ('MOVMSKPD', '660F50D6', 0x40, 'movmskpd edx,xmm6', 'movmskpd edx,xmm6'),
    ('PEXTRD', '660F3A16D011', 0x40, 'pextrd eax,xmm2,17', 'pextrd eax,xmm2,17'),
    ('PEXTRD 2', '660F3A161011', 0x40, 'pextrd dword [eax],xmm2,17', 'pextrd dword [eax],xmm2,17'),
    ('PEXTRD', '660F3A16EA11', 0x40, 'pextrd edx,xmm5,17', 'pextrd edx,xmm5,17'),
    ('PEXTRD 2', '660F3A161011', 0x40, 'pextrd dword [eax],xmm2,17', 'pextrd dword [eax],xmm2,17'),
    ('MOVBE',   '0F38F04910', 0x40, 'movbe ecx,dword [ecx + 16]', 'movbe ecx,dword [ecx + 16]'),
    ('MOVBE 2', '0F38F009', 0x40, 'movbe ecx,dword [ecx]', 'movbe ecx,dword [ecx]'),
    ('MOVBE 3', '0F38F00C8D41414141', 0x40, 'movbe ecx,dword [0x41414141 + ecx * 4]', 'movbe ecx,dword [0x41414141 + ecx * 4]'),
    ('MOVBE 4', '0F38F15110', 0x40, 'movbe dword [ecx + 16],edx', 'movbe dword [ecx + 16],edx'),
    ('MOVBE 5', '0F38F102', 0x40, 'movbe dword [edx],eax', 'movbe dword [edx],eax'),
    ('MOVBE 6', '0F38F10541414141', 0x40, 'movbe dword [0x41414141],eax', 'movbe dword [0x41414141],eax'),
    ('MOVBE 7', '660F38F00541414141', 0x40, 'movbe ax,word [0x41414141]', 'movbe ax,word [0x41414141]'),
    # TODO: so technically this is movhlps, but we need to have the operand change the opcode
    ('MOVLPS', '0F12DE', 0x40, 'movlps xmm3,xmm6', 'movlps xmm3,xmm6'),
    ('MOVLPS 2', '0F121D41414141', 0x40, 'movlps xmm3,qword [0x41414141]', 'movlps xmm3,qword [0x41414141]'),
    ('SUBPD', '660F5C6C240E', 0x40, 'subpd xmm5,oword [esp + 14]', 'subpd xmm5,oword [esp + 14]'),
    ('MAXPD', '660F5FAB0F270000', 0x40, 'maxpd xmm5,oword [ebx + 9999]', 'maxpd xmm5,oword [ebx + 9999]'),
    ('XORPD', '660F57BD15CD5B07', 0x40, 'xorpd xmm7,oword [ebp + 123456789]', 'xorpd xmm7,oword [ebp + 123456789]'),
    ('SQRTPD', '660f51ca', 0x40, 'sqrtpd xmm1,xmm2', 'sqrtpd xmm1,xmm2'),
    ('PSHUFD', '660F70CD11', 0x40, 'pshufd xmm1,xmm5,17', 'pshufd xmm1,xmm5,17'),
    ('PEXTRW', '660FC5C307', 0x40, 'pextrw eax,xmm3,7', 'pextrw eax,xmm3,7'),
    ('MOVQ', '660FD620', 0x40, 'movq qword [eax],xmm4', 'movq qword [eax],xmm4'),
    ('PMAXUB', '660FDE2541414141', 0x40, 'pmaxub xmm4,oword [0x41414141]', 'pmaxub xmm4,oword [0x41414141]'),
    ('MOVNTDQ', '660FE73D78563412', 0x40, 'movntdq oword [0x12345678],xmm7', 'movntdq oword [0x12345678],xmm7'),
    ('PADDD', '660FFECE', 0x40, 'paddd xmm1,xmm6', 'paddd xmm1,xmm6'),
    ('VPCEXT', '0F3F070B', 0x40, 'vpcext 7,11','vpcext 7,11'),

    ('MOV AMETH_C', '0f20d0', 0x40, 'mov eax,ctrl2', 'mov eax,ctrl2'),
    ('MOV AMETH_C 2', '0f20f1', 0x40, 'mov ecx,ctrl6', 'mov ecx,ctrl6'),
    ('MOV AMETH_C 3', '0f22e2', 0x40, 'mov ctrl4,edx', 'mov ctrl4,edx'),
    ('MOV AMETH_C 4', '0f22f8', 0x40, 'mov ctrl7,eax', 'mov ctrl7,eax'),

    ('MOV AMETH_D', '0f21c0', 0x40, 'mov eax,debug0', 'mov eax,debug0'),
    ('MOV AMETH_D 2', '0f21f9', 0x40, 'mov ecx,debug7', 'mov ecx,debug7'),
    ('MOV AMETH_D 3', '0f23e1', 0x40, 'mov debug4,ecx', 'mov debug4,ecx'),
    ('PMOVMSKB', '660fd7f8', 0x40, 'pmovmskb edi,xmm0', 'pmovmskb edi,xmm0'),
    ('PMOVMSBK 2', '660fd7ca', 0x40, 'pmovmskb ecx,xmm2', 'pmovmskb ecx,xmm2'),
    ('PMOVMSKB 3', '0fd7f8', 0x40, 'pmovmskb edi,mm0', 'pmovmskb edi,mm0'),
    ('MOV SEGREG', '64a130000000', 0x40, 'fs: mov eax,dword [0x00000030]', 'fs: mov eax,dword [0x00000030]'),
    ('MOV SEGREG 2', '8ce0', 0x40, 'mov eax,fs', 'mov eax,fs'),
    ('MOV SEGREG 3', '8ec6', 0x40, 'mov es,esi', 'mov es,esi'),
    ('MOV SEGREG 4', '668ec6', 0x40, 'mov es,si', 'mov es,si'),
    ('MOV SEGREG 6', '8E142541414141', 0x40, 'mov ss,word [0x41414141]', 'mov ss,word [0x41414141]'),
    ('MOV SEGREG 7', '8C042541414141', 0x40, 'mov word [0x41414141],es', 'mov word [0x41414141],es'),

    ('LEA', '8d5a0c', 0x40, 'lea ebx,dword [edx + 12]', 'lea ebx,dword [edx + 12]'),
    ('SIGNED', '83C0F9', 0x40, 'add eax,0xfffffff9', 'add eax,0xfffffff9'),
    ('MAXPD', '660F5F64C020', 0x40, 'maxpd xmm4,oword [eax + eax * 8 + 32]', 'maxpd xmm4,oword [eax + eax * 8 + 32]'),
    ('MAXPD 2', '660f5fa490d0a80000', 0x40, 'maxpd xmm4,oword [eax + edx * 4 + 43216]', 'maxpd xmm4,oword [eax + edx * 4 + 43216]'),
    # ('rm4mod2', '', 0x40, '', ''),

    ('PMOVSXBW', '660f3820ca', 0x40, 'pmovsxbw xmm1,xmm2', 'pmovsxbw xmm1,xmm2'),
    ('PMOVSXBD', '660f3821cb', 0x40, 'pmovsxbd xmm1,xmm3', 'pmovsxbd xmm1,xmm3'),
    ('PMOVSXBQ', '660f3822d3', 0x40, 'pmovsxbq xmm2,xmm3', 'pmovsxbq xmm2,xmm3'),
    ('PMOVSXWD', '660f3823ff', 0x40, 'pmovsxwd xmm7,xmm7', 'pmovsxwd xmm7,xmm7'),
    ('PMOVSXWQ', '660f3824dc', 0x40, 'pmovsxwq xmm3,xmm4', 'pmovsxwq xmm3,xmm4'),
    ('PMOVSXDQ', '660f3825df', 0x40, 'pmovsxdq xmm3,xmm7', 'pmovsxdq xmm3,xmm7'),

    ('PMOVZXBW', '660f3830ca', 0x40, 'pmovzxbw xmm1,xmm2', 'pmovzxbw xmm1,xmm2'),
    ('PMOVZXBD', '660f3831cb', 0x40, 'pmovzxbd xmm1,xmm3', 'pmovzxbd xmm1,xmm3'),
    ('PMOVZXBQ', '660f3832d3', 0x40, 'pmovzxbq xmm2,xmm3', 'pmovzxbq xmm2,xmm3'),
    ('PMOVZXWD', '660f3833ff', 0x40, 'pmovzxwd xmm7,xmm7', 'pmovzxwd xmm7,xmm7'),
    ('PMOVZXWQ', '660f3834dc', 0x40, 'pmovzxwq xmm3,xmm4', 'pmovzxwq xmm3,xmm4'),
    ('PMOVZXDQ', '660f3835df', 0x40, 'pmovzxdq xmm3,xmm7', 'pmovzxdq xmm3,xmm7'),

    ('PMOVSXBW (MEM)', '660f382018', 0x40, 'pmovsxbw xmm3,qword [eax]', 'pmovsxbw xmm3,qword [eax]'),
    ('PMOVSXBD (MEM)', '660f3821242541414141', 0x40, 'pmovsxbd xmm4,dword [0x41414141]', 'pmovsxbd xmm4,dword [0x41414141]'),
    ('PMOVSXBQ (MEM)', '660f38228b29230000', 0x40, 'pmovsxbq xmm1,word [ebx + 9001]', 'pmovsxbq xmm1,word [ebx + 9001]'),
    ('PMOVSXWD (MEM)', '660f38230c11', 0x40, 'pmovsxwd xmm1,qword [ecx + edx]', 'pmovsxwd xmm1,qword [ecx + edx]'),
    ('PMOVSXWQ (MEM)', '660f38241cb507000000', 0x40, 'pmovsxwq xmm3,dword [0x00000007 + esi * 4]', 'pmovsxwq xmm3,dword [0x00000007 + esi * 4]'),
    ('PMOVSXDQ (MEM)', '660f382532', 0x40, 'pmovsxdq xmm6,qword [edx]', 'pmovsxdq xmm6,qword [edx]'),

    ('PMOVSXBW (MEM)', '660f383018', 0x40, 'pmovzxbw xmm3,qword [eax]', 'pmovzxbw xmm3,qword [eax]'),
    ('PMOVSXBD (MEM)', '660f3831242541414141', 0x40, 'pmovzxbd xmm4,dword [0x41414141]', 'pmovzxbd xmm4,dword [0x41414141]'),
    ('PMOVSXBQ (MEM)', '660f38328b29230000', 0x40, 'pmovzxbq xmm1,word [ebx + 9001]', 'pmovzxbq xmm1,word [ebx + 9001]'),
    ('PMOVSXWD (MEM)', '660f38330c11', 0x40, 'pmovzxwd xmm1,qword [ecx + edx]', 'pmovzxwd xmm1,qword [ecx + edx]'),
    ('PMOVSXWQ (MEM)', '660f38341cb507000000', 0x40, 'pmovzxwq xmm3,dword [0x00000007 + esi * 4]', 'pmovzxwq xmm3,dword [0x00000007 + esi * 4]'),
    ('PMOVSXDQ (MEM)', '660f383532', 0x40, 'pmovzxdq xmm6,qword [edx]', 'pmovzxdq xmm6,qword [edx]'),
    ('SFENCE', '0faef8', 0x40, 'sfence ', 'sfence '),
    ('LFENCE', '0faee8', 0x40, 'lfence ', 'lfence '),
    ('MFENCE', '0faef0', 0x40, 'mfence ', 'mfence '),
    ('XSAVE', '0FAE2541414141', 0x40, 'xsave dword [0x41414141]', 'xsave dword [0x41414141]'),

    ('ptest', '660F3817242541414141', 0x40, 'ptest xmm4,oword [0x41414141]', 'ptest xmm4,oword [0x41414141]'),
    # TODO: Should we add the implicit xmm0 to these? It's only in non-vex mode
    ('blendvps', '660F3814CB', 0x40, 'blendvps xmm1,xmm3', 'blendvps xmm1,xmm3'),
    ('pblendvb', '660F3815CB', 0x40, 'blendvpd xmm1,xmm3', 'blendvpd xmm1,xmm3'),

    # AES-NI feature set
    ('AESENC', '660F38DCEA', 0x40, 'aesenc xmm5,xmm2', 'aesenc xmm5,xmm2'),
    ('AESENC (MEM)', '660f38DC3A', 0x40, 'aesenc xmm7,oword [edx]', 'aesenc xmm7,oword [edx]'),
    ('AESENC (MEM 2)', '660f38DC7C2404', 0x40, 'aesenc xmm7,oword [esp + 4]', 'aesenc xmm7,oword [esp + 4]'),
    ('AESENC (MEM 3)', '660F38DC1D41414141', 0x40, 'aesenc xmm3,oword [0x41414141]', 'aesenc xmm3,oword [0x41414141]'),
    ('AESENCLAST', '660F38DDDC', 0x40, 'aesenclast xmm3,xmm4', 'aesenclast xmm3,xmm4'),
    ('AESENCLAST (MEM)', '660F38DD18', 0x40, 'aesenclast xmm3,oword [eax]', 'aesenclast xmm3,oword [eax]'),
    ('AESENCLAST (MEM 2)', '660F38DD5808', 0x40, 'aesenclast xmm3,oword [eax + 8]', 'aesenclast xmm3,oword [eax + 8]'),
    ('AESENCLAST (MEM 3)', '660F38DD2578563442', 0x40, 'aesenclast xmm4,oword [0x42345678]', 'aesenclast xmm4,oword [0x42345678]'),
    ('AESDEC', '660f38DECB', 0x40, 'aesdec xmm1,xmm3', 'aesdec xmm1,xmm3'),
    ('AESDEC (MEM)', '660F38DE0C24', 0x40, 'aesdec xmm1,oword [esp]', 'aesdec xmm1,oword [esp]'),
    ('AESDEC (MEM 2)', '660F38DE5D0C', 0x40, 'aesdec xmm3,oword [ebp + 12]', 'aesdec xmm3,oword [ebp + 12]'),
    ('AESDEC (MEM 3)', '660F38DE3544434241', 0x40, 'aesdec xmm6,oword [0x41424344]', 'aesdec xmm6,oword [0x41424344]'),
    ('AESDECLAST', '660F38DFED', 0x40, 'aesdeclast xmm5,xmm5', 'aesdeclast xmm5,xmm5'),
    ('AESDECLAST (MEM)', '660F38DF2E', 0x40, 'aesdeclast xmm5,oword [esi]', 'aesdeclast xmm5,oword [esi]'),
    ('AESDECLAST (MEM 2)', '660F38DF6740', 0x40, 'aesdeclast xmm4,oword [edi + 64]', 'aesdeclast xmm4,oword [edi + 64]'),
    ('AESDECLAST (MEM 3)', '660F38DF2511213141', 0x40, 'aesdeclast xmm4,oword [0x41312111]', 'aesdeclast xmm4,oword [0x41312111]'),
    ('AESIMC', '660F38DBF9', 0x40, 'aesimc xmm7,xmm1', 'aesimc xmm7,xmm1'),
    ('AESIMC (MEM)', '660F38DB13', 0x40, 'aesimc xmm2,oword [ebx]', 'aesimc xmm2,oword [ebx]'),
    ('AESIMC (MEM 2)', '660F38DB5020', 0x40, 'aesimc xmm2,oword [eax + 32]', 'aesimc xmm2,oword [eax + 32]'),
    ('AESIMC (MEM 3)', '660F38DB1D00000041', 0x40, 'aesimc xmm3,oword [0x41000000]', 'aesimc xmm3,oword [0x41000000]'),
    ('AESKEYGENASSIST', '660F3ADFFE08', 0x40, 'aeskeygenassist xmm7,xmm6,8', 'aeskeygenassist xmm7,xmm6,8'),
    ('AESKEYGENASSIST (MEM)', '660F3ADF1AFE', 0x40, 'aeskeygenassist xmm3,oword [edx],254', 'aeskeygenassist xmm3,oword [edx],254'),
    ('AESKEYGENASSIST (MEM 2)', '660F3ADF998000000039', 0x40, 'aeskeygenassist xmm3,oword [ecx + 128],57', 'aeskeygenassist xmm3,oword [ecx + 128],57'),
    ('AESKEYGENASSIST (MEM 3)', '660F3ADF2541414141C6', 0x40, 'aeskeygenassist xmm4,oword [0x41414141],198', 'aeskeygenassist xmm4,oword [0x41414141],198'),
    ('PCLMULQDQ', '660F3A44D307', 0x40, 'pclmulqdq xmm2,xmm3,7', 'pclmulqdq xmm2,xmm3,7'),
    ('PCLMULQDQ (MEM)', '660F3A441007', 0x40, 'pclmulqdq xmm2,oword [eax],7', 'pclmulqdq xmm2,oword [eax],7'),
    ('PCLMULQDQ (MEM 2)', '660F3A4478119C', 0x40, 'pclmulqdq xmm7,oword [eax + 17],156', 'pclmulqdq xmm7,oword [eax + 17],156'),
    ('PCLMULQDQ (MEM 3)', '660F3A443D41414141C6', 0x40, 'pclmulqdq xmm7,oword [0x41414141],198', 'pclmulqdq xmm7,oword [0x41414141],198'),
    ('RDRAND', '0fc7f0', 0x40, 'rdrand eax', 'rdrand eax'),
    ('RDSEED', '0fc7f8', 0x40, 'rdseed eax', 'rdseed eax'),
    ('VMPTRST', '0fc73d41414141', 0x40, 'vmptrst qword [0x41414141]', 'vmptrst qword [0x41414141]'),
    ('VMCLEAR', '0fc73541414141', 0x40, 'vmptrld qword [0x41414141]', 'vmptrld qword [0x41414141]'),
    ('CMPXCHG', '0fb0d0', 0x40, 'cmpxchg al,dl', 'cmpxchg al,dl'),
    ('ADOX', 'F30F38f6c2', 0x40, 'adox eax,edx', 'adox eax,edx'),
    ('ADOX MEM', 'F30F38F6042541414141', 0x40, 'adox eax,dword [0x41414141]', 'adox eax,dword [0x41414141]'),
    ('ADCX', '660f38f6e5', 0x40, 'adcx esp,ebp', 'adcx esp,ebp'),
    ('ADCX MEM', '660f38f6242541414141', 0x40, 'adcx esp,dword [0x41414141]', 'adcx esp,dword [0x41414141]'),
]


class i386InstructionSet(unittest.TestCase):
    _arch = envi.getArchModule("i386")

    def check_opreprs(self, opcodes):
        vw = vivisect.VivWorkspace()
        scanv = e_memcanvas.StringMemoryCanvas(vw)

        for name, bytez, va, reprOp, renderOp in opcodes:
            try:
                op = self._arch.archParseOpcode(binascii.unhexlify(bytez), 0, va)
            except envi.InvalidInstruction:
                self.fail("Failed to parse opcode bytes: %s (case: %s, expected: %s)" % (bytez, name, reprOp))
            except Exception:
                self.fail("Failed to parse opcode bytes: %s (case: %s, expected: %s)" % (bytez, name, reprOp))
            msg = '%s failed length check. Got %d, expected %d' % (name, len(op), int(len(bytez)/2))
            self.assertEqual(len(op), int(len(bytez)/2), msg=msg)
            # print("'%s', 0x%x, '%s' == '%s'" % (bytez, va, repr(op), reprOp))
            try:
                self.assertEqual(repr(op), reprOp)
            except AssertionError:
                self.fail("Failing match for case %s (%s != %s)" % (name, repr(op), reprOp))

            scanv.clearCanvas()
            op.render(scanv)
            self.assertEqual(scanv.strval, renderOp)

    def test_envi_i386_disasm_Specific_SingleByte_Instrs(self):
        self.check_opreprs(i386SingleByteOpcodes)

    def test_envi_i386_disasm_Specific_MultiByte_Instrs(self):
        self.check_opreprs(i386MultiByteOpcodes)

    def checkOpcode(self, hexbytez, va, oprepr, opcheck, opercheck, renderOp):

        op = self._arch.archParseOpcode(binascii.unhexlify(hexbytez), 0, va)
        self.assertEqual(repr(op), oprepr)
        opvars = vars(op)
        for opk, opv in opcheck.items():
            self.assertEqual((opk, opvars.get(opk)), (opk, opv))

        for oidx in range(len(op.opers)):
            oper = op.opers[oidx]
            opervars = vars(oper)
            for opk, opv in opercheck[oidx].items():
                self.assertEqual((opk, opervars.get(opk)), (opk, opv))

        vw = vivisect.VivWorkspace()
        scanv = e_memcanvas.StringMemoryCanvas(vw)
        op.render(scanv)
        self.assertEqual(scanv.strval, renderOp)

    def test_envi_i386_disasm_Reg_Operands(self):
        '''
        test an opcode encoded with an Reg operand
        _0      add al      04
        G       add         02
        C       mov         0f20
        D       mov         0f21
        P       punpcklbw   0f60
        S       mov         8c
        U       movmskps    0f50
        V       sqrtps      0f51

        '''
        opbytez = '0032'
        oprepr = 'add byte [edx],dh'
        opcheck = {'iflags': 65536, 'va': 16384, 'repr': None, 'prefixes': 0, 'mnem': 'add', 'opcode': 8193, 'size': 2}
        opercheck = ({'disp': 0, 'tsize': 1, '_is_deref': True, 'reg': 2}, {'tsize': 1, 'reg': 134742018},)
        self.checkOpcode(opbytez, 0x4000, oprepr, opcheck, opercheck, oprepr)

        opbytez = '0440'
        oprepr = 'add al,64'
        opcheck = {'iflags': 65536, 'prefixes': 0, 'mnem': 'add', 'opcode': 8193, 'size': 2}
        opercheck = ({'tsize': 1, 'reg': 524288}, {'tsize': 1, 'imm': 64})
        self.checkOpcode(opbytez, 0x4000, oprepr, opcheck, opercheck, oprepr)

        opbytez = '0218'
        oprepr = 'add bl,byte [eax]'
        opcheck = {'iflags': 65536, 'va': 16384, 'repr': None, 'prefixes': 0, 'mnem': 'add', 'opcode': 8193, 'size': 2}
        opercheck = ( {'tsize': 1, 'reg': 524291}, {'disp': 0, 'tsize': 1, '_is_deref': True, 'reg': 0} )
        self.checkOpcode( opbytez, 0x4000, oprepr, opcheck, opercheck, oprepr )

        opbytez = '0f2018'
        oprepr = 'mov dword [eax],ctrl3'
        opcheck =  {'iflags': 65536, 'va': 16384, 'repr': None, 'prefixes': 0, 'mnem': 'mov', 'opcode': 24577, 'size': 3}
        opercheck = ( {'disp': 0, 'tsize': 4, '_is_deref': True, 'reg': 0}, {'tsize': 4, 'reg': 27} )
        self.checkOpcode( opbytez, 0x4000, oprepr, opcheck, opercheck, oprepr )

    def test_envi_i386_disasm_Imm_Operands(self):
        '''
        test an opcode encoded with an Imm operand
        _0      rol         d000
        A       callf       9a
        '''
        opbytez = 'd000'
        oprepr = 'rol byte [eax],1'
        opcheck =  {'iflags': 65536, 'va': 16384, 'prefixes': 0, 'mnem': 'rol', 'opcode': 8201, 'size': 2}
        opercheck = ( {'disp': 0, 'tsize': 1, '_is_deref': True, 'reg': 0}, {'tsize': 4, 'imm': 1} )
        self.checkOpcode( opbytez, 0x4000, oprepr, opcheck, opercheck, oprepr )

        # this is failing legitimately... we decode this opcode wrong
        opbytez = '9aaa11aabbcc33'
        oprepr = 'callf 0x33cc:0xbbaa11aa'
        opcheck =  {'iflags': 65540, 'va': 16384, 'repr': None, 'prefixes': 0, 'mnem': 'callf', 'opcode': 4099, 'size': 7}
        opercheck = [{'tsize': 6, 'imm': 56954414829994}]
        self.checkOpcode( opbytez, 0x4000, oprepr, opcheck, opercheck, oprepr )

    def test_envi_i386_disasm_PcRel_Operands(self):
        '''
        test an opcode encoded with an PcRelative operand
        '''
        pass

    def test_envi_i386_disasm_RegMem_Operands(self):
        '''
        test an opcode encoded with an RegMem operand
        X       outsb       6e
        Y       insd        6d
        '''
        opbytez = '6e'
        oprepr = 'outsb edx,byte [esi]'
        opcheck = {'iflags': 65538, 'va': 16384, 'repr': None, 'prefixes': 0, 'mnem': 'outsb', 'opcode': 57347, 'size': 1}
        opercheck = [{'tsize': 4, 'reg': 2}, {'disp': 0, 'tsize': 1, '_is_deref': True, 'reg': 6}]
        self.checkOpcode(opbytez, 0x4000, oprepr, opcheck, opercheck, oprepr)

        opbytez = '6d'
        oprepr = 'insd dword [esi],edx'
        opcheck =  {'iflags': 65538, 'va': 16384, 'repr': None, 'prefixes': 0, 'mnem': 'insd', 'opcode': 57346, 'size': 1}
        opercheck = [{'disp': 0, 'tsize': 4, '_is_deref': True, 'reg': 6}, {'tsize': 4, 'reg': 2}]
        self.checkOpcode( opbytez, 0x4000, oprepr, opcheck, opercheck, oprepr )

    def test_envi_i386_disasm_ImmMem_Operands(self):
        '''
        test an opcode encoded with an ImmMem operand
        O       mov         a1
        '''
        opbytez = 'a1a2345678'
        oprepr = 'mov eax,dword [0x785634a2]'
        opcheck =  {'iflags': 65536, 'va': 16384, 'repr': None, 'prefixes': 0, 'mnem': 'mov', 'opcode': 24577, 'size': 5}
        opercheck = [{'tsize': 4, 'reg': 0}, {'tsize': 4, '_is_deref': True, 'imm': 2018915490}]
        self.checkOpcode( opbytez, 0x4000, oprepr, opcheck, opercheck, oprepr )

    def test_envi_i386_disasm_SIB_Operands(self):
        '''
        exercize the entire SIB operand space
        A       jmp         ea
        E       lar         0f02
        Q       cvttps2pi   0f2c
        W       cvttps2pi   0f2c
        '''
        opbytez = 'eaa123456789ab'          # this wants more bytes, why?
        oprepr = 'jmp 0xab89:0x674523a1'       # this repr's wrong.  it should be ab89:674523a1
        opcheck = {'iflags': 65545, 'va': 16384, 'repr': None, 'prefixes': 0, 'mnem': 'jmp', 'opcode': 4097, 'size': 7}
        opercheck = [{'tsize': 6, 'imm': 188606631453601}]
        self.checkOpcode(opbytez, 0x4000, oprepr, opcheck, opercheck, oprepr)

        opbytez = '0f02aabbccddee'
        oprepr = 'lar ebp,word [edx - 287454021]'
        opcheck = {'iflags': 65538, 'va': 16384, 'repr': None, 'prefixes': 0, 'mnem': 'lar', 'opcode': 57344, 'size': 7}
        opercheck = [{'tsize': 4, 'reg': 5}, {'disp': -287454021, 'tsize': 2, '_is_deref': True, 'reg': 2}]
        self.checkOpcode( opbytez, 0x4000, oprepr, opcheck, opercheck, oprepr )

        opbytez = '0f2caabbccddeeff'
        oprepr = 'cvttps2pi mm5,qword [edx - 287454021]'
        opcheck = {'iflags': 65536, 'va': 16384, 'repr': None, 'prefixes': 0, 'mnem': 'cvttps2pi', 'opcode': 61440}
        opercheck = [{'tsize': 8, 'reg': 4194355}, {'disp': -287454021, 'tsize': 8, '_is_deref': True, 'reg': 2}]
        self.checkOpcode(opbytez, 0x4000, oprepr, opcheck, opercheck, oprepr)



import envi
import envi.exc as e_exc
import envi.bits as e_bits

import logging

from vivisect.const import *

logger = logging.getLogger(__name__)
BRANCH_FLAGS = envi.IF_BRANCH | envi.IF_CALL


class EmulationMonitor:
    """
    Emulation monitors may be passed into functions like
    runFunction() to track and hook the emulator.
    """
    def __init__(self):
        # FIXME make this a dict and re-plumb for VaSet
        self.emuanom = []  # A list of emulation anomalies (in va,msg tuples)
        self.retvals = []  # A list of the return values seen

    def logAnomaly(self, emu, va, msg):
        self.emuanom.append((va, msg))

    def getAnomalies(self):
        return list(self.emuanom)

    def prehook(self, emu, op, starteip):
        """
        This monitor hook gets called back prior to the execution of
        each instruction in the emulator.
        """
        pass

    def posthook(self, emu, op, endeip):
        """
        This monitor hook gets called back following the execution of
        each instruction in the emulator. During this callback, the
        emulator's curpath variable remains on the previous instruction.
        """
        pass

    def apicall(self, emu, op, pc, api, argv):
        '''
        This monitor hook gets called when the workspace emulator has
        encountered a call to another subroutine.  The API definition
        ( see vivisect/impapi ) as well as the parsed arguments are
        provided.

        Return a value other than None to to set the return value for
        the called API.
        '''
        pass

class AnalysisMonitor(EmulationMonitor):
    '''
    A shared class for the various arch emulation implementations
    which contains utility functions for some standard enumeration.
    '''
    def __init__(self, vw, fva):
        EmulationMonitor.__init__(self)
        self.vw = vw
        self.fva = fva
        self.onceop = {}
        self.stackmax = 0
        self.stackargs = {}
        self.operrefs = []
        self.callcomments = []
        self._dynamic_branch_handlers = []

    def addAnalysisResults(self, vw, emu):
        '''
        Do any post-run annotation that the base analysis emulator knows
        how to do...
        '''
        # Add emulation anomalies
        for row in self.getAnomalies():
            va, msg = row
            vw.setVaSetRow("Emulation Anomalies", row)
            vw.setComment(va, 'Emu Anomaly: %s' % (msg,), check=True)

        # Go through the evaluated dereference operands and add operand refs
        deltadone = {}
        for va, idx, val, tsize, spdelta, discrete in self.operrefs:
            if spdelta:
                vw.addFref(self.fva, va, idx, spdelta)

                if deltadone.get(spdelta):
                    continue

                deltadone[spdelta] = True
                if spdelta <= 0:  # add function locals
                    vw.setFunctionLocal(self.fva, spdelta, LSYM_NAME, ('int','local%d' % abs(spdelta)))

                continue

            # Only infer things about the workspace based on discrete operands
            if vw.isValidPointer(val) and discrete:
                vw.addXref(va, val, REF_DATA)
                if vw.getLocation(val) is not None:
                    continue

                vw.guessDataPointer(val, tsize)

        imports = self.vw.getImports()
        for va, callname, argv in self.callcomments:
            reprargs = [emu.reprVivValue(val) for val in argv]
            self.vw.setComment(va, '%s(%s)' % (callname, ','.join(reprargs)))
            cva = self.vw.vaByName(callname)
            if not cva:
                # Look for imports that match the call name.
                for import_va, _, _, import_name in imports:
                    if import_name == callname:
                        cva = import_va
                        break
            if cva:
                self.vw.addXref(va, cva, REF_CODE, envi.BR_PROC)

    def addDynamicBranchHandler(self, cb):
        '''
        Add a callback handler for dynamic branches the code-flow resolver
        doesn't know what to do with
        '''
        if cb in self._dynamic_branch_handlers:
            raise Exception("Already have this handler (%s) for dynamic branches" % repr(cb))

        self._dynamic_branch_handlers.append(cb)

    def logAnomaly(self, emu, eip, msg):
        self.vw.vprint("EmuAnom: 0x%.8x (f:0x%.8x) %s" % (eip, self.fva, msg))
        return EmulationMonitor.logAnomaly(self, self, eip, msg)

    def prehook(self, emu, op, starteip):

        if not self.onceop.get(starteip):
            self.onceop[starteip] = True
            for i, o in enumerate(op.opers):
                if o.isDeref():
                    discrete = o.isDiscrete()
                    operva = o.getOperAddr(op, emu)
                    # keep track of the max here, but save it for later too...
                    stackoff = emu.getStackOffset(operva)
                    if stackoff and stackoff >= 0:
                        self.stackmax = max(self.stackmax, stackoff)
                        self.stackargs[stackoff] = True

                    self.operrefs.append((starteip, i, operva, o.tsize, stackoff, discrete))

        if op.iflags & BRANCH_FLAGS:
            oper = op.opers[0]
            if oper.isDeref() or oper.isReg():
                for cb in self._dynamic_branch_handlers:
                    try:
                        cb(self, emu, op, starteip)
                    except Exception as e:
                        logger.exception('error with dyn branch handler (%s) (%s)', cb, e)

    def apicall(self, emu, op, pc, api, argv):
        rettype, retname, convname, callname, callargs = api
        if self.vw.getComment(op.va) is None:
            if callname is None:
                callname = self.vw.getName(pc)

            self.callcomments.append((op.va, callname, argv))

        # Record uninitialized register use
        for i, arg in enumerate(argv):

            # Check for taints first because it's faster...
            taint = emu.getVivTaint(arg)
            if taint:
                tva, ttype, tinfo = taint
                if ttype == 'uninitreg':
                    emu.logUninitRegUse(tinfo)
                continue

            # Lets see if the API def has type info for us...
            if self.vw.isValidPointer(arg):
                argtype, argname = callargs[i]
                self.vw.setComment(arg, argtype, check=True)
                if not self.vw.isLocation(arg):
                    if argname == 'funcptr':
                        logger.debug('discovered new function: 0x%x', arg)
                        self.vw.makeFunction(arg)

                    # FIXME make an API for this? ( the name parsing )
                    # Normalize and guess about the structure...
                    typeguess = argtype.strip().strip('*').split()
                    if typeguess[0] == 'struct' and len(typeguess) >= 2:
                        vs = self.vw.getStructure(arg, typeguess[1])
                        if vs is not None:
                            self.vw.makeStructure(arg, typeguess[1], vs=vs)

                continue

        # From here down, we are only checking for instances where the
        # emulator detected a call, and managed to resolve some code flow
        # that we failed to recognize statically...
        if pc == self.fva:
            self.vw.setFunctionMeta(self.fva, 'Recursive', True)
            return

        if self.vw.isFunction(pc):
            return

        # Ditch "call 0" constructs...
        if pc == op.va + len(op):
            return

        if not self.vw.isExecutable(pc):
            return

        # WOOT - we have found a runtime resolved function!
        self.vw.vprint('0x%.8x: Emulation Found 0x%.8x (from func: 0x%.8x) via %s' % (op.va, pc, self.fva, repr(op)))
        self.vw.makeFunction(pc, arch=op.iflags & envi.ARCH_MASK)
        self.vw.addXref(op.va, pc, REF_CODE, envi.BR_PROC)
